from lolpop.pipeline.train.base_train import BaseTrain
from lolpop.utils import common_utils as utils

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class OfflineTrain(BaseTrain): 
    __REQUIRED_CONF__ = {
        "component": ["data_splitter", "metadata_tracker", "resource_version_control", "model_checker", "model_explainer", "model_visualizer", "model_bias_checker"], 
        "config": []
    }

    def split_data(self, data, *args, **kwargs): 
        """
        Split the input data into train/validation/test sets using the data splitter configured 
        in the class. 
        
        Args:
            data (object): A DataFrame containing the input data to be split.

        Returns:
            A dictionary with train/test/validation data where each is a 
            DataFrame containing the corresponding split.
        """

        #split data. returns dictionary of train/vali/test dataframes
        data_dict = self.data_splitter.split_data(data)
        
        return data_dict

    def train_model(self, data, *args, **kwargs): 
        """
        Train a machine learning model using the input data and return the trained model and model version. 
        If a hyperparameter tuner is configured in the class, the tuner will be used to search for the optimal 
        hyperparameters for the model.
        
        Args:
            data (object): A dictionary containing three data splits "train_data", "val_data" and "test_data", as 
                  returned by the split_data method.
            *args: Additional non-keyword arguments.
            **kwargs: Optional keyword arguments to pass to the model trainer.

        Returns:
            A tuple of the trained model and model version.
        """

        #create model_version
        id = self._get_config("model_name")
        model_version = self.metadata_tracker.create_resource(id, type="model_version")
    
        #if we are using hyperparameter tuner, use that, otherwise just use the model trainer provided
        if hasattr(self, "hyperparameter_tuner"): 
            model = self.hyperparameter_tuner.run_experiment(data, model_version)
        else:
            model, _  = self.model_trainer.build_model(data, model_version)

        return model, model_version

    def check_model(self, data_dict, model, model_version, *args, **kwargs):
        """
        Perform model checks on the input data and trained model, log the results in the 
        metadata tracker, and notify via email if any issues are found with the checks. 
        
        Args:
            data_dict (object): A dictionary containing split data, as 
                       returned by the split_data method.
            model (object): The trained machine learning model to be checked.
            model_version (object): The version of the model being checked.
            **kwargs: Optional keyword arguments to pass to the model checker.
        """

        #run model checks
        model_report, file_path, checks_status = self.model_checker.check_model(data_dict, model)

        #log model report to metadata tracker
        self.metadata_tracker.log_checks(
            model_version,
            file_path = file_path, 
            report = model_report, 
            checker_class = type(self.model_checker).__name__, 
            type="model"
            )

        if checks_status == "ERROR" or checks_status == "WARN": 
            url = self.metadata_tracker.url
            self.notify("Issues found with model checks. Visit %s for more information." %url, checks_status)


    def analyze_model(self, data_dict, model, model_version, *args, **kwargs): 
        """
        Generate feature importance plots, compare the model and data to a baseline, 
        and generate visualizations to assist in model analysis. 
        
        Args:
            data_dict (object): A dictionary containing three data splits, as 
                       returned by the split_data method.
            model (object): The trained machine learning model to be analyzed.
            model_version (object): The version of the model to be analyzed.
            **kwargs: Optional keyword arguments to pass to the model explainer, baseline comparison, and 
                      model visualizer.
        """
        transformed_data = model._transform_dict(data_dict)

        #calculate feature importance
        self.model_explainer.get_feature_importance(
            transformed_data, model, model_version)
         
        #compare model to baseline
        self.model_checker.get_baseline_comparison(
            data_dict, model, model_version)

        #create some eye candy
        self.model_visualizer.generate_viz(
            transformed_data, model._get_model(), model_version)


    def compare_models(self, data_dict, model, model_version, *args, **kwargs): 
        """
        Compare the current model to the previously deployed model, log the comparison results in the 
        metadata tracker, and return a boolean indicating if the current model is an improvement over 
        the previous model. 
        
        Args:
            data_dict (object): A dictionary containing three data splits, as 
                       returned by the split_data method.
            model (object): The trained machine learning model to be analyzed.
            model_version (object): The version of the model being tested for improvement.
            
        Returns:
            A boolean indicating if the current model is an improvement over the previous model. 
            If there is no previous model to compare to, will return True.
        """
        #get currently deployed model version/previous model version and winning experiment
        prev_model_version = self.metadata_tracker.get_currently_deployed_model_version(model_version)

        if prev_model_version is not None: 
            prev_experiment = self.metadata_tracker.get_winning_experiment(prev_model_version)

            #get the prev model version object from resource version control and load it into a model trainer
            prev_model_obj = self.resource_version_control.get_model(prev_experiment)
            prev_model = self.metadata_tracker.load_model(prev_model_obj, model_version, model)

            #compare model to the last version 
            report, file_path = self.model_checker.calculate_model_drift(data_dict, model, prev_model)
            self.metadata_tracker.log_artifact(model_version, id="model_drift_report", path = file_path, external = False)
            is_new_model_better = self.model_checker.compare_models(data_dict, model, prev_model, model_version, prev_model_version)
            
            return is_new_model_better 
        
        else: 
            self.log("No previous model version found for model: %s" %self.metadata_tracker.get_resource_id(model_version))
            return True 

    def check_model_bias(self, data_dict, model, model_version, *args, **kwargs):
        """
        Check for bias in the input data and trained model and log the results in the metadata tracker. 
        
        Args:
            data_dict (object): A dictionary containing three data splits, as 
                       returned by the split_data method.
            model (object): The trained machine learning model to be checked for bias.
            model_version (object): The version of the model to be checked for bias.
        """
        #check model bias
        bias_metrics = self.model_bias_checker.check_model_bias(data_dict, model)
        for k,v in bias_metrics.items(): 
            self.metrics_tracker.log_metric(model_version, k, v)

    def retrain_model_on_all_data(self, data, model_version, ref_model = None):
        """
        Retrain a machine learning model on all available data and return the new trained model and model version. 
        
        Args:
            data (object): A DataFrame containing all the data to be used for retraining.
            model_version (object): The version of the model to be retrained.
            ref_model (object): The previous model to be used as reference.

        Returns:
            A tuple of the trained model and model version.
        """
        if hasattr(self, "model_trainer"):  
            model, experiment = self.model_trainer.rebuild_model(data, model_version)
        else: 
            #if no model_trainer is specified in config, we can rebuild using the winning_exp_model_trainer
            winning_exp_model = self.metadata_tracker.load_model(None, model_version, ref_model=ref_model)
            model, experiment = winning_exp_model.rebuild_model(data, model_version)
            return model, experiment



from lolpop.component.model_checker.base_model_checker import BaseModelChecker
from lolpop.utils import common_utils as utils
from deepchecks.tabular import Dataset
from deepchecks.tabular.suites import model_evaluation
from deepchecks.tabular.checks import TrainTestLabelDrift

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class DeepchecksModelChecker(BaseModelChecker): 

    __REQUIRED_CONF__ = {
        "config": ["local_dir"]
    }

    __DEFAULT_CONF__ = {
        "config": {"DEEPCHECKS_MODEL_REPORT_NAME": "DEEPCHECKS_MODEL_REPORT.HTML",
                   "DEEPCHECKS_MODEL_DRIFT_REPORT_NAME": "DEEPCHECKS_MODEL_DRIFT_REPORT.HTML"}
    }

    def check_model(self, data_dict, model, **kwargs): 
        """
        Checks the model against data_dict. Saves the Deepchecks model suite report in HTML format.

        Args:
        data_dict: dictionary, the dictionary containing data for the model
        model: object, the object of the model class

        Returns:
        model_report: object, the model suite report
        file_path: str, the location of the saved HTML file
        checks_status: str, the status (PASS, ERROR or WARN) of model checks
        """
        model_target = self._get_config("MODEL_TARGET")
        model_index = self._get_config("MODEL_INDEX")
        model_time_index = self._get_config("MODEL_TIME_INDEX")
        model_cat_features = self._get_config("MODEL_CAT_FEATURES")

        #set up data + predictions for train/text drift
        df_train, df_test = self.data_splitter.get_train_test_dfs(data_dict) 

        ds_train = Dataset(df_train, label = model_target, index_name=model_index, cat_features=model_cat_features, datetime_name=model_time_index)
        ds_test = Dataset(df_test, label = model_target, index_name=model_index, cat_features=model_cat_features, datetime_name=model_time_index)
      
        model_suite = model_evaluation() 
        model_report = model_suite.run(ds_train, ds_test, model._get_model())

        file_path = "%s/%s" %(self._get_config("local_dir"), self._get_config("DEEPCHECKS_REPORT_NAME"))
        model_report.save_as_html(file_path)
        
        checks_status = "PASS"
        if len(model_report.get_not_passed_checks()) > 0: 
            checks_status = "ERROR"
        elif len(model_report.get_not_ran_checks()) > 0: 
            checks_status = "WARN"

        return model_report, file_path, checks_status

    def calculate_model_drift(self, data, current_model, deployed_model, **kwargs): 
        """
        Compares the performance of current_model and deployed_model on data. Saves the model drift report in HTML format.

        Args:
        data: dictionary, the dictionary containing data for the models
        current_model: object, the current state model object
        deployed_model: object, the already deployed model object

        Returns:
        model_report: object, the train/test drift report object
        file_path: str, the location of the saved HTML file
        """
        model_target = self._get_config("MODEL_TARGET")
        model_index = self._get_config("MODEL_INDEX")
        model_time_index = self._get_config("MODEL_TIME_INDEX")
        model_cat_features = self._get_config("MODEL_CAT_FEATURES")

        df_current = data["X_test"].copy() 
        df_deployed = df_current.copy()
        df_current["prediction"] = current_model.transform_and_predict_df(df_current)
        df_deployed["prediction"] = deployed_model.transform_and_predict_df(df_deployed)

        ds_current = Dataset(df_current, label = "prediction", index_name=model_index, cat_features=model_cat_features, datetime_name=model_time_index)
        ds_deployed = Dataset(df_deployed, label = "prediction", index_name=model_index, cat_features=model_cat_features, datetime_name=model_time_index)
      
        check = TrainTestLabelDrift()
        model_report = check.run(train_dataset=ds_deployed, test_dataset=ds_current)

        file_path = "%s/%s" % (self._get_config("local_dir"),
                               self._get_config("DEEPCHECKS_MODEL_DRIFT_REPORT_NAME"))
        model_report.save_as_html(file_path)

        return model_report, file_path




from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

import boto3
from tqdm import tqdm

import pandas as pd
import io
import os


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class S3DataConnector(BaseDataConnector):
    __REQUIRED_CONF__ = {"config": ["AWS_S3_BUCKET",
                                    "AWS_ACCESS_KEY_ID",
                                    "AWS_SECRET_KEY_ID",
                                    "AWS_SESSION_TOKEN", ]}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aws_config = utils.load_config(
            [
                "AWS_S3_BUCKET", 
                "AWS_ACCESS_KEY_ID", 
                "AWS_SECRET_KEY_ID", 
                "AWS_SESSION_TOKEN",
            ]
        )

    def get_data(self, path, *args, **kwargs):
        """
        Reads data from the AWS S3 Bucket.

        Args:
            path: a string representing the location of the file in the bucket

        Returns:
            DataFrame containing data from the bucket.
        """
        df = self._load_data(path, self.aws_config)
        return df

    def save_data(self, data, path, *args, **kwargs):
        """
        Saves data into the S3 Bucket.

        Args:
            data: Data to be written into the bucket
            path: a string representing the location of the file in the bucket

        Returns:
            None
        """
        self._save_data(data, path, self.aws_config)


    @classmethod
    def _load_data(self, path, config, *args, **kwargs):
        """
        Reads files within the AWS S3 Bucket.

        Args:
            path: a string representing the location of the file in the bucket
            config: a dictionary that stores AWS S3 Buckets, AWS Access Key ID, AWS Secret Access Key, and AWS Session Token.


        Returns:
            Data on file.
        """
        client = self._get_client(config)
        bucket = config.get("AWS_S3_BUCKET")
        extension = path.split(".")[-1]

        response = client.get_object(Bucket=bucket, Key=path)
        status = response.get("ResponseMetadata", {}).get("HTTPStatusCode")

        if status == 200:
            self.log(f"Successful S3 get_object response. Status - {status}")
            source_data = response.get("Body").read()
            if extension == "csv":
                data = pd.read_csv(io.StringIO(source_data),
                                encoding="utf-8", **kwargs)
            elif extension == "parquet" or extension == "pq":
                data = pd.read_parquet(io.BytesIO(source_data), **kwargs)
            elif extension == "orc":
                data = pd.read_orc(io.BytesIO(source_data), **kwargs)
            else:
                self.log("Unsupported file type provided: %s" % extension)
            self.log("Successfully loaded data from %s into DataFrame." % path)

        else:
            self.log(f"Unsuccessful S3 get_object response. Status - {status}")

        return data

    def _save_data(self, data, path, config, *args, **kwargs):
        """
        Saves Data into AWS S3 Bucket

        Args:
            data: Data to be written into the bucket
            path: a string representing the location of the file in the bucket
            config: a dictionary that stores AWS S3 Buckets, AWS Access Key ID, AWS Secret Access Key, and AWS Session Token.

        Returns:
            None
        """
        # get client
        client = self._get_client(config)
        bucket = config.get("AWS_S3_BUCKET")
        extension = path.split(".")[-1]

        if extension == ".csv":
            f = io.StringIO()
            data.to_csv(f, index=False)
        elif extension == ".pq" or extension == "parquet":
            f = io.BytesIO()
            data.to_parquet(f, index=False)
        else:
            raise Exception("Unsupported file type for path: %s" %(path))
        f.seek(0)

        response = client.put_object(
            Bucket=bucket, Key=path, Body=f.getvalue()
        )

        status = response.get("ResponseMetadata", {}).get("HTTPStatusCode")

        if status == 200:
            self.log(f"Successful S3 put_object response. Status - {status}")
        else:
            self.log(f"Unsuccessful S3 put_object response. Status - {status}")


    def _get_client(self, config):
        """
        Returns the AWS S3 client.

        Args:
            config: a dictionary that stores AWS S3 Buckets, AWS Access Key ID, AWS Secret Access Key, and AWS Session Token.

        Returns:
            An AWS S3 client.
        """
        client = boto3.client(
            "s3",
            aws_access_key_id=config.get("AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=config.get("AWS_SECRET_ACCESS_KEY"),
            aws_session_token=config.get("AWS_SESSION_TOKEN"),
        )
 
        return client



from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils
from lolpop.component.data_connector.snowflake_data_connector import SnowflakeDataConnector

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class SnowflakeDataTransformer(BaseDataTransformer): 

    __REQUIRED_CONF__ = {"config": ["SNOWFLAKE_ACCOUNT",
                                    "SNOWFLAKE_USER",
                                    "SNOWFLAKE_PASSWORD",
                                    "SNOWFLAKE_DATABASE",
                                    "SNOWFLAKE_SCHEMA",
                                    "SNOWFLAKE_WAREHOUSE"]}

    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs)
        self.snowflake_config = utils.load_config([
            "SNOWFLAKE_ACCOUNT", 
            "SNOWFLAKE_USER", 
            "SNOWFLAKE_PASSWORD", 
            "SNOWFLAKE_DATABASE", 
            "SNOWFLAKE_SCHEMA", 
            "SNOWFLAKE_WAREHOUSE"], self.config)

    def transform(self, sql, *args, **kwargs):
        """
        Transforms data from Snowflake database.

        Args:
            sql (str): The SQL statement used to extract data from Snowflake database.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            data: The data extracted from Snowflake database.

        """
        data = SnowflakeDataConnector._load_data(sql, self.snowflake_config)
        return data 



from lolpop.component.model_checker.base_model_checker import BaseModelChecker
from lolpop.utils import common_utils as utils
from evidently.test_suite import TestSuite
from evidently.tests import TestColumnDrift
from evidently.report import Report
from evidently.test_preset import MulticlassClassificationTestPreset, NoTargetPerformanceTestPreset, BinaryClassificationTestPreset, RegressionTestPreset
from evidently.metric_preset import DataDriftPreset, TargetDriftPreset, DataQualityPreset
from evidently import ColumnMapping

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class EvidentlyAIModelChecker(BaseModelChecker): 

    __REQUIRED_CONF__ = {
        "config": ["local_dir", "model_target"]
    }

    __DEFAULT_CONF__ = {
        "config": {"EVIDENTLYAI_MODEL_REPORT_NAME": "EVIDENTLYAI_MODEL_REPORT.HTML",
                   "EVIDENTLYAI_MODEL_DRIFT_REPORT_NAME": "EVIDENTLYAI_MODEL_DRIFT_REPORT.HTML"}
    }

    def check_model(self, data_dict, model, *args, **kwargs):
        """This class is used to check and calculate drift for a trained machine learning model using the EvidentlyAI testing framework. This class inherits BaseModelChecker. 
    
    Methods:
    -----------
    check_model(data_dict, model, **kwargs)
        Runs varies model test presents using EvidentlyAI testing framework.

        Parameters:
        -----------
        data_dict : dict
            A dictionary containing training and testing data in the form of pandas dataframes.
        model : object
            A trained machine learning model.
        **kwargs : Arbitrary keyword arguments

        Returns:
        --------
        model_report : object
            A TestSuite Object containing results of model drift tests.
        file_path : str
            The path where the EVIDENTLY_MODEL_REPORT.HTML is stored.
        checks_status : str
            The status of model drift test. It can be "ERROR", "WARN" or "PASS".
        """
        if self.problem_type == "classification": 
            classification_type = utils.get_multiclass(data_dict["y_train"].unique())

        #set up column mapping 
        column_mapping = ColumnMapping()
        model_target = self._get_config("MODEL_TARGET")
        column_mapping.target = model_target
        column_mapping.prediction = "prediction"
        column_mapping.id = self._get_config("model_index")
        column_mapping.datetime = self._get_config("model_time_index")

        #set up data + predictions for train/test drift
        df_train, df_test = self.data_splitter.get_train_test_dfs(data_dict) 
        df_train["prediction"] = model.transform_and_predict_df(df_train.drop([model_target], axis=1))
        df_test["prediction"] = model.transform_and_predict_df(df_test.drop([model_target], axis=1))

        if self.problem_type == "classification": 
            if classification_type == "multiclass": 
                model_report = TestSuite(tests=[MulticlassClassificationTestPreset(), NoTargetPerformanceTestPreset(), TestColumnDrift(column_name=model_target)])
            else: 
                model_report = TestSuite(tests=[BinaryClassificationTestPreset(), NoTargetPerformanceTestPreset(), TestColumnDrift(column_name=model_target)])
        elif self.problem_type == "regression":
            model_report = TestSuite(tests=[RegressionTestPreset(), NoTargetPerformanceTestPreset(), TestColumnDrift(column_name=model_target)])
        else: 
            self.notify("Unsupported problem type: %s" %self.problem_type)
        model_report.run(current_data=df_test, reference_data=df_train, column_mapping=column_mapping)
        file_path = "%s/%s" %(self._get_config("local_dir"), self._get_config("evidentlyai_model_report_name"))
        model_report.save_html(file_path)

        summary = model_report.as_dict()["summary"]

        checks_status = "PASS"
        if summary["failed_tests"] > 0: 
            checks_status = "ERROR"
        elif summary["success_tests"] < summary["total_tests"]: 
            checks_status = "WARN"

        return  model_report, file_path, checks_status

    def calculate_model_drift(self, data, current_model, deployed_model, *args, **kwargs):
        """Calculate the drift between two trained machine learning models using EvidentlyAI testing framework.

        Parameters:
        -----------
        data : dict
            A dictionary containing training and testing data in the form of pandas dataframes.
        current_model : object
            A trained machine learning model.
        deployed_model : object
            A trained machine learning model.

        Returns:
        --------
        drift_report : object
            A TestSuite Object containing results of model drift tests.
        file_path : str
            The path where the EVIDENTLY_MODEL_DRIFT_REPORT.HTML is stored.
        """
        #set up dfs
        df_current = data["X_test"].copy() 
        df_deployed = df_current.copy()

        #create column mapping
        column_mapping = ColumnMapping()
        column_mapping.target = check_col_exists(
            self._get_config("MODEL_TARGET"), df_current, df_deployed)
        column_mapping.prediction = "prediction"
        column_mapping.id = check_col_exists(
            self._get_config("model_index"), df_current, df_deployed)
        column_mapping.datetime = check_col_exists(
            self._get_config("model_time_index"), df_current, df_deployed)


        df_current["prediction"] = current_model.transform_and_predict_df(df_current)
        df_deployed["prediction"] = deployed_model.transform_and_predict_df(df_deployed)

        #drift_report = TestSuite(tests=[TestColumnDrift(column_name="prediction")])
        drift_report = Report(metrics = [TargetDriftPreset()])

        drift_report.run(current_data=df_current, reference_data=df_deployed, column_mapping=column_mapping)
        file_path = "%s/%s" % (self._get_config(
            "local_dir"), self._get_config("evidentlyai_model_drift_report_name"))
        drift_report.save_html(file_path)

        return drift_report, file_path
        
#drift reports error if one dataset doesn't contain something in the column mapping
# so this ensures we don't get into a mapping that will cause an error.


def check_col_exists(col, dfA, dfB):
    if col in dfA.columns and col in dfB.columns:
        return col
    else:
        return None



from lolpop.component.data_checker.base_data_checker import BaseDataChecker
from lolpop.utils import common_utils as utils
from deepchecks.tabular import Dataset
from deepchecks.tabular.suites import data_integrity
import os 

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class DeepchecksDataChecker(BaseDataChecker): 
    
    __REQUIRED_CONF__ = {
        "config" : ["local_dir"]
    }

    __DEFAULT_CONF__ = {
        "config": {"DEEPCHECKS_REPORT_NAME":"DEEPCHECKS_DATA_REPORT.HTML"}
    }

    def check_data(self, data, *args, **kwargs):
        """Generates a data check report using Deepchecks.

        Args:
            data (pd.DataFrame): A dataframe of the data to check

        Returns:
            data_report (object): Python object of the data report.
            file_path (string):  Path to the exported report. 
            checks_status (string): Status of the checks ("PASS"/"WARN"/"ERROR", etc.)
        """        
        model_target = self._get_config("MODEL_TARGET")
        label = None 
        if model_target in data.columns:
            label = model_target
            
        ds = Dataset(
            data, 
            label = label, 
            index_name=self._get_config("MODEL_INDEX"), 
            cat_features=self._get_config("MODEL_CAT_FEATURES"), 
            datetime_name=self._get_config("MODEL_TIME_INDEX")
            )
        data_suite = data_integrity() 
        data_report = data_suite.run(ds)
        file_path = "%s/%s.html" % (self._get_config("local_dir"), self._get_config("DEEPCHECKS_REPORT_NAME"))
        #deepchecks doesn't ovewrite existing files, so we need to remove this path if it already exists
        try:
            os.remove(file_path)
        except: 
            pass
        data_report.save_as_html(file_path)

        num_checks_passed = len(data_report.get_passed_checks())
        num_checks_failed = len(data_report.get_not_passed_checks())
        num_checks_not_ran = len(data_report.get_not_ran_checks())
        self.log("%s had %s passed checks." %(self.name, num_checks_passed))
        self.log("%s had %s failed checks." % (self.name, num_checks_failed))
        self.log("%s had %s checks not run." % (self.name, num_checks_not_ran))

        checks_status = "PASS"
        if num_checks_failed > 0:
            checks_status = "ERROR"
        elif num_checks_not_ran > 0: 
            checks_status = "WARN"

        return data_report, file_path, checks_status



from lolpop.component.base_component import BaseComponent

class BaseTestRecorder(BaseComponent):

    test_results = []

    def record_test(self, obj, method, test, test_method, result, *args, **kwargs):
        pass

    def print_report(*args, **kwargs): 
        pass 

    def generate_report(*args, **kwargs): 
        pass 



from lolpop.base_integration import BaseIntegration

class BaseComponent(BaseIntegration): 

    def __init__(self, *args, **kwargs):
        super().__init__(integration_type="component", *args, **kwargs)


from lolpop.pipeline.deploy.base_deploy import BaseDeploy
from lolpop.utils import common_utils as utils
from lolpop.utils import metaflow_utils as meta_utils
from metaflow import FlowSpec, step
from pathlib import Path

METAFLOW_CLASS = "MetaflowOfflineDeploySpec"
PLUGIN_PATHS = "plugin_paths.txt"

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class MetaflowOfflineDeploy(BaseDeploy):

    __REQUIRED_CONF__ = {
        "component": ["metadata_tracker", "model_repository", "resource_version_control", "model_deployer"]
    }

    def run(self, model, model_version, *args, **kwargs):
        """
        Runs the MetaflowOfflineDeploy.

        This method loads a Metaflow flow object based on the current file, loads the flow, 
        and then runs it using the model and version provided. This method logs execution 
        and returns nothing. 

        Parameters:
        -----------
            model : str
                The name of the model to run.
            model_version : str
                The version of the model to run.
            **kwargs : dict
                Other, non-specified parameters 

        Returns:
        --------

        Raises:
        -------
        """
        #get flow class object from this file
        mod_cl = meta_utils.get_flow_class(__file__, METAFLOW_CLASS)

        flow = meta_utils.load_flow(
            mod_cl, self, PLUGIN_PATHS, model=model, model_version=model_version)
        self.log("Loaded metaflow flow %s" % METAFLOW_CLASS)

        meta_utils.run_flow(flow, "run", __file__, PLUGIN_PATHS)
        self.log("Metaflow pipeline %s finished." % METAFLOW_CLASS)

    def get_artifacts(self, artifact_keys):
        """
        Obtains artifacts of the MetaflowOfflineDeploy.

        This method gets the latest run of the MetaflowOfflineDeploy pipeline and returns the 
        requested artifacts. 

        Parameters:
        -----------
            artifact_keys : list
                A list of keys for the artifacts being requested. 

        Returns:
        --------
            artifacts : dict
                A dictionary of requested artifacts.
        
        Raises:
        -------
        """
        #get latest run of this pipeline
        run = meta_utils.get_latest_run(METAFLOW_CLASS)

        #get requested artifacts
        artifacts = meta_utils.get_run_artifacts(
            run, artifact_keys, METAFLOW_CLASS)

        return artifacts


class MetaflowOfflineDeploySpec(FlowSpec):

    def __init__(self, lolpop=None, model=None, model_version=None, use_cli=False, **kwargs):
        #you need to set local attributes before calling super
        #only the first time we create the class will these parameters be provided.
        #The rest of the calls are metaflow internal calls and we do not want to reset them
        if lolpop is not None:
            self.lolpop = lolpop
        if model is not None:
            self.model = model
        if model_version is not None: 
            self.model_version = model_version

        #load plugins
        meta_utils.load_plugins(PLUGIN_PATHS)

        #call init -- this kicks off the main workflow
        super().__init__(use_cli=use_cli)

    @step
    def start(self):
        self.next(self.promote_model)

    @step
    def promote_model(self):
        #get model from RVC if not provided.
        if self.model is None:
            if self.model_version is not None:
                experiment = self.lolpop.metadata_tracker.get_winning_experiment(
                    self.model_version)
                model_obj = self.lolpop.resource_version_control.get_model(experiment)
                self.model = self.lolpop.metadata_tracker.load_model(
                    model_obj, self.model_version, ref_model=None)
            else:
                self.lolpop.notify(
                    "Must provide either model_version or model in order to promote a model.")

        #register model in model repository
        model_id = self.lolpop.model_repository.register_model(self.model_version, self.model)

        #promote model
        promotion = self.lolpop.model_repository.promote_model(model_id)
        self.promotion = promotion 

        self.next(self.check_approval)

    @step
    def check_approval(self):
        is_approved = self.lolpop.model_repository.check_approval(self.promotion)

        if not is_approved and self.lolpop._get_config("auto_approve_models"):
            self.lolpop.model_repository.approve_model(self.promotion)
            is_approved = True

        self.is_approved = is_approved

        self.next(self.deploy_model)

    @step
    def deploy_model(self):
        if self.is_approved: 
            deployment = self.lolpop.model_deployer.deploy_model(self.promotion, self.model_version)
            self.deployment = deployment

        self.next(self.end)

    @step
    def end(self):
        pass
    

if __name__ == '__main__':
    #need to use CLI here because metaflow calls back into this function to
    #execute each individual step
    MetaflowOfflineDeploySpec(use_cli=True)



from lolpop.pipeline.base_pipeline import BasePipeline
from typing import Any 

class BaseProcess(BasePipeline): 

    datasets_used = []

    def __init__(self, pipeline_type = "process", *args, **kwargs):
        super().__init__(pipeline_type=pipeline_type, *args, **kwargs)

    def transform_data(self, source, *args, **kwargs) -> Any: 
        pass 

    def track_data(self, data, id, *args, **kwargs) -> Any: 
        pass

    def profile_data(self, data, dataset_version, *args, **kwargs): 
        pass 

    def check_data(self, data, dataset_version, *args, **kwargs): 
        pass 

    def compare_data(self, data, dataset_version, *args, **kwargs): 
        pass 




from lolpop.component.model_visualizer.base_model_visualizer import BaseModelVisualizer
from lolpop.utils import common_utils as utils
from matplotlib import pyplot as plt 
from prophet.diagnostics import cross_validation, performance_metrics
from prophet.plot import plot_cross_validation_metric

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class ProphetModelVisualizer(BaseModelVisualizer):

    __REQUIRED_CONF__ = {
        "component": ["metadata_tracker"],
    }

    def generate_viz(self, data, model, model_version, forecast_period=7, forecast_frequency="D", *args, **kwargs):
        """ Method to generate visualizations of Prophet model components and forecasts. 
    
        Args:
        
            data: pandas DataFrame
                Input DataFrame containing time-series data.
            model: Prophet model object
                A trained Prophet model object.
            model_version: object
                Model version from the metadata_tracker
            forecast_period: int, default 7
                The period for which forecast is to be generated.
            forecast_frequency: str, default 'D'
                The frequency of the forecast period.
        
        Returns:
        
            None
        """
        # generate component plot
        id = self.metadata_tracker.get_resource_id(model_version)
        historic_predictions = model.predict(model.history)
        model.plot_components(historic_predictions)
        local_dir = self._get_config("local_dir")
        self._save_and_log_file(local_dir, id, "component_plot.png", model_version)

        #plot forecast
        future = model.make_future_dataframe(
            periods=forecast_period,
            freq=forecast_frequency,
            include_history=True
        )
        predictions = model.predict(future)
        model.plot(predictions)
        self._save_and_log_file(local_dir, id, "forecast_plot.png", model_version)

    def cross_validation(self, model, model_version, initial, period, horizon, *args, **kwargs): 
        """Method to perform time-series cross validation on the Prophet model.
    
        Args:
        
            model: Prophet model object
                A trained Prophet model object.
            model_version: str
                String defining the model version.
            initial: str, int or float
                String or numerical value to define the size of the initial training period.
            period: str, int or float
                String or numerical value to define the length of the spacing between cutoff dates.
            horizon: str, int or float
                String or numerical value to define the forecast horizon.
        
        Returns:
        
            None
        """

        id = self.metadata_tracker.get_resource_id(model_version)
        df_cv = cross_validation(
            model,
            initial=initial,
            period=period,
            horizon=horizon,
        )

        df_p = performance_metrics(df_cv)

        #save performance metrics
        local_dir = self._get_config("local_dir")
        file_path = "%s/%s_%s" %(local_dir, id, "cv_performance_metrics.csv")
        df_p.to_csv(file_path)
        self.metadata_tracker.log_artifact(model_version, "cv_performance_metrics", file_path)

        #save plots
        for metric in df_p.drop(["horizon"], axis=1, errors="ignore"): 
            try: 
                plot_cross_validation_metric(df_cv, metric=metric)
                self._save_and_log_file(local_dir, id, "cv_performance_%s.png" % metric, model_version)
            except: 
                pass 
                

        
    def _save_and_log_file(self, dir, id, file_name, model_version): 
        """Helper method to save plots generated by ProphetModelVisualizer methods to local directory and log them with the metadata tracker.
    
        Args:
        
            dir: str
                Directory path (String) for saving the file.
            id: str
                Unique identifier for the model version.
            file_name: str
                Name for the file to be saved.
            model_version: object
                model version to log file to 
        
        Returns:
        
            None
        
        """
        file_path = "%s/%s_%s" % (dir, id, file_name)
        plt.savefig(file_path)
        name = file_name.split(".")[0]
        if model_version is not None: 
            self.metadata_tracker.log_artifact(model_version, name, file_path)
        plt.close()



from lolpop.component.feature_transformer.base_feature_transformer import BaseFeatureTransformer
from lolpop.utils import common_utils as utils

from sklearn.pipeline import Pipeline
from feature_engine import imputation, encoding, discretisation, outliers, transformation, creation, datetime, selection, preprocessing
from feature_engine.timeseries import forecasting   


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class FeatureEngineFeatureTransformer(BaseFeatureTransformer):
    __REQUIRED_CONF__ = {"config": ["transformers"]
                         }

    __DEFAULT_CONF__ = {
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        #get transformer config and process
        transformers = self._get_config("transformers",[])
        i = 1
        transformer_array = []
        for t in transformers: 
            transformer = self._get_transformer(
                t.get("transformer"),
                list(t.get("transformer_columns")),
                *t.get("transformer_args",[]), 
                **t.get("transformer_kwargs",{})
                )
            name = "transform_%s" %i 
            i+=1 
            transformer_array.append((name, transformer))
        
        #now we can call ColumnTranformer on the transformer_array 
        self.params = self._get_config("pipeline_kwargs",{})
        self.transformer = Pipeline(transformer_array, **self.params)


    def fit(self, X_data, y_data=None, *args, **kwargs):
        """
        Fit the feature transformer to the data.

        Args:
            X_data (object): The input data to fit the transformer to.
            y_data (object): The target data.
            *args: Positional arguments to be passed to the `fit` method.
            **kwargs: Keyword arguments to be passed to the `fit` method.

        Returns:
            ColumnTransformer: The fitted ColumnTransformer object.

        Raises:
            Exception: If no feature transformer is found.

        """
        if not hasattr(self, "transformer"):
            raise Exception(
                "No feature transformer found. Unable to fit transformer.")
        self.transformer = self.transformer.fit(X_data, y_data, **kwargs)

        return self.transformer

    def transform(self, data, *args, **kwargs):
        """
        Apply the fitted feature transformer to transform the input data.

        Args:
            data: The input data to be transformed.
            *args: Positional arguments to be passed to the `transform` method.
            **kwargs: Keyword arguments to be passed to the `transform` method.

        Returns:
            numpy.ndarray or scipy.sparse matrix: The transformed data.

        Raises:
            Exception: If no feature transformer is found.

        """
        if not hasattr(self, "transformer"):
            raise Exception(
                "No feature transformer found. Unable to transform features.")

        data_out = self.transformer.transform(data, **kwargs)

        return data_out
    
    def fit_transform(self, X_data, y_data=None, *args, **kwargs): 
        """
        Fit the feature transformer to the data and transform it.

        Args:
            X_data: The input data to fit the transformer to and transform.
            y_data: The target data. 
            *args: Positional arguments to be passed to the `fit_transform` method.
            **kwargs: Keyword arguments to be passed to the `fit_transform` method.

        Returns:
            numpy.ndarray or scipy.sparse matrix: The transformed data.

        Raises:
            Exception: If no feature transformer is found.

        """
        if not hasattr(self, "transformer"):
            raise Exception(
                "No feature transformer found. Unable to transform features.")

        data_out = self.transformer.fit_transform(X_data, y_data, **kwargs)

        return data_out

    def _get_transformer(self, transformer_class, columns, *args, **kwargs): 
        """
        Get the feature_engine transformer object based on the given class name.

        Args:
            transformer_class: The class name of the transformer.
            columns: list of columns to apply the transformer to 
            *args: Positional arguments to be passed to the transformer constructor.
            **kwargs: Keyword arguments to be passed to the transformer constructor.

        Returns:
            object: The instantiated scikit-learn transformer object.

        Raises:
            Exception: If there is an error loading the transformer from `sklearn.preprocessing`.

        """
        try: 
            transformer = self._map_transformer(transformer_class) 
            return transformer(variables=columns, *args, **kwargs)
        except Exception as e: 
            raise Exception("Error loading transformer %s from feature_engine." %transformer_class)
        
    def _map_transformer(self, transformer_class): 
        """
        Map the transformer class name to the corresponding feature_engine transformer object.

        Args:
            transformer_class (str): The class name of the transformer.

        Returns:
            object: The instantiated feature_engine transformer object.

        Raises:
            None

        """
        imputer_list = [
            "MeanMedianImputer", 
            "ArbitraryNumberImputer", 
            "EndTailImputer",
            "CategoricalImputer", 
            "RandomSampleImputer",
            "AddMissingIndicator",
            "DropMissingData"
        ]
        
        encoder_list = [
            "OneHotEncoder",
            "CountFrequencyEncoder",
            "OrdinalEncoder",
            "MeanEncoder",
            "WoEEncoder",
            "DecisionTreeEncoder",
            "RareLabelEncoder",
            "StringSimilarityEncoder",
        ]

        discretiser_list = [
            "ArbitraryDiscretiser",
            "EqualFrequencyDiscretiser",
            "EqualWidthDiscretiser",
            "DecisionTreeDiscretiser",
            "GeometricWidthDiscretiser",
        ]

        outlier_list = [
            "ArbitraryOutlierCapper",
            "Winsorizer",
            "OutlierTrimmer",
        ]

        numerical_list = [
            "LogTransformer",
            "LogCpTransformer",
            "ReciprocalTransformer",
            "PowerTransformer",
            "BoxCoxTransformer",
            "YeoJohnsonTransformer",
            "ArcsinTransformer",
        ]

        creation_list = [
            "MathFeatures",
            "RelativeFeatures",
            "CyclicalFeatures",
        ]

        datetime_list = [
            "DatetimeFeatures",
            "DatetimeSubtraction",
        ]

        selection_list = [
            "DropFeatures",
            "DropConstantFeatures",
            "DropDuplicateFeatures",
            "DropCorrelatedFeatures",
            "SmartCorrelatedSelection",
            "DropHighPSIFeatures",
            "SelectByInformationValue",
            "SelectByShuffling",
            "SelectBySingleFeaturePerformance",
            "SelectByTargetMeanPerformance",
            "RecursiveFeatureElimination",
            "RecursiveFeatureAddition",
            "ProbeFeatureSelection",
        ]

        forecasting_list = [
            "LagFeatures",
            "WindowFeatures",
            "ExpandingWindowFeatures",
        ]

        preprocessing_list = [
            "MatchCategories",
            "MatchVariables",
        ]

        transformer = None 

        if transformer_class in imputer_list:
            transformer = getattr(imputation, transformer_class)

        elif transformer_class in encoder_list: 
            transformer = getattr(encoding, transformer_class)

        elif transformer_class in discretiser_list: 
            transformer = getattr(discretisation, transformer_class)
        
        elif transformer_class in outlier_list: 
            transformer = getattr(outliers, transformer_class)

        elif transformer_class in numerical_list: 
            transformer = getattr(transformation, transformer_class)

        elif transformer_class in creation_list: 
            transformer = getattr(creation, transformer_class)

        elif transformer_class in datetime_list: 
            transformer = getattr(datetime, transformer_class)

        elif transformer_class in selection_list: 
            transformer = getattr(selection, transformer_class)

        elif transformer_class in forecasting_list:  
            transformer = getattr(forecasting, transformer_class)

        elif transformer_class in preprocessing_list: 
            transformer = getattr(preprocessing, transformer_class)

        else: 
            self.log("Transformer class %s not found!" %transformer_class, level="WARN")

        return transformer




from lolpop.component.base_component import BaseComponent
from typing import Any

class BaseGenAIChatbot(BaseComponent):

    def ask(self, prompt, *args, **kwargs) -> str:
        pass

    def prepare_message(self, role, content, *args, **kwargs) -> dict[str,Any]: 
        pass


from lolpop.component.model_bias_checker.base_model_bias_checker import BaseModelBiasChecker
from lolpop.utils import common_utils as utils
from aif360.datasets import StandardDataset
from aif360.metrics import BinaryLabelDatasetMetric, ClassificationMetric
from omegaconf.listconfig import ListConfig
@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class AIFairnessModelBiasChecker(BaseModelBiasChecker): 

    __REQUIRED_CONF__ = {"config": ["model_target",
                                   "privileged_groups", "unprivileged_groups", 
                                   "favorable_classes", "privileged_classes", 
                                   "protected_attribute_names"]}

    def check_model_bias(self, data, model, *args, **kwargs): 
        """Uses the provided data and model to check for bias in the model's predictions.
            Args:
                data (dict): A dictionary containing the training/testing data.
                model (object): The model trainer object being checked for bias.
        """
        metrics_out = {}
        if self.problem_type == "classification": 
            #set up data/predictions
            df_train = data["X_train"]
            df_train_w_labels = df_train.copy() 
            df_train_w_labels[self._get_config("model_target")] = data["y_train"].values

            #set up aif dataset + configs
            model_target = self._get_config("model_target")
            privileged_groups = self._get_config("privileged_groups")
            unprivileged_groups = self._get_config("unprivileged_groups")
            favorable_classes = self._eval_classes(self._get_config("favorable_classes"))
            privileged_classes = self._eval_classes(self._get_config("privileged_classes"))
            ds = StandardDataset(
                df=df_train_w_labels.dropna(), #NA make AI360 fail 
                label_name=model_target, 
                protected_attribute_names=self._get_config("protected_attribute_names"), 
                favorable_classes=favorable_classes, 
                privileged_classes=privileged_classes,
                )

            #get fairness metrics from binarylabeldatasetmetric. This essentially checks bias in the dataset itself
            bldm = BinaryLabelDatasetMetric(ds, unprivileged_groups=unprivileged_groups, privileged_groups=privileged_groups)
            metrics_out["bldm_base_rate"] = bldm.base_rate() # ratio of priv/total 
            metrics_out["bldm_disparate_impact"] = bldm.disparate_impact() # ratio of prob fav outcomes of non/priv
            metrics_out["bldm_mean_difference"] =  bldm.mean_difference() # diff in probability of favorable outcoems for non vs priv
            metrics_out["bldm_consistency"] = bldm.consistency()[0] # meastures how similar the labels are for similar instanes
            metrics_out["bldm_smoothed_edf"] = bldm.smoothed_empirical_differential_fairness() #

            ds_preds = ds.copy()
            ds_preds.labels = model.transform_and_predict(data).get("train")

            #get fairness metric from classificationmetric. This checks fairness in predictions
            cm = ClassificationMetric(ds, ds_preds, unprivileged_groups=unprivileged_groups, privileged_groups=privileged_groups)
            metrics_out["cm_tpr_diff"] = cm.true_positive_rate_difference() 
            metrics_out["cm_selection_rate"] = cm.selection_rate() 
            metrics_out["cm_theil_index"] = cm.theil_index() 
            metrics_out["cm_false_positive_rate_ratio"] = cm.false_positive_rate_ratio() 
            metrics_out["cm_false_omission_rate_ratio"] = cm.false_omission_rate_ratio() 
            metrics_out["cm_false_negative_rate_ratio"] = cm.false_negative_rate_ratio()
            metrics_out["cm_false_discovery_rate_ratio"] = cm.false_discovery_rate_ratio() 
            metrics_out["cm_error_rate_ratio"] = cm.error_rate_ratio()  
            metrics_out["cm_differential_fba"] = cm.differential_fairness_bias_amplification()
        else: 
            self.log("Problem type %s not supported for model bias checker %s" %(self.problem_type, self.name))        

        #TODO: we should have bias thresholds, and if any are exceeded, then we can reweigh the dataset
        #via something like: https://aif360.readthedocs.io/en/stable/modules/generated/aif360.algorithms.preprocessing.Reweighing.html
        #will have to think about how this applies in the workflow though.
        #i.e. we may want to move the dataset bias stuff up in the data_processing step. 

        return metrics_out

    def _eval_classes(self, class_list): 
        """Evalutes `lambda` functions defined in configuration into proper python functions. 
           AIF360 allows dynamically specifying classes via lambda notation, but this isn't 
           parsed correctly when defined in yaml files. This function fixes that. 

        Args:
            class_list (list): List of classes

        Returns:
            list_out: list of classes
        """
        list_out = []
        for cl in class_list:
            if isinstance(cl, str) and  cl.startswith("lambda x"): 
                cl = eval(cl) #we should rework this in the future 
            elif isinstance(cl, list) or isinstance(cl, ListConfig): 
                cl = self._eval_classes(cl)
            list_out.append(cl)  
        return list_out      


from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

import pandas as pd
from google.cloud import bigquery
from google.oauth2 import service_account

import os 

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class BigQueryDataConnector(BaseDataConnector):
    __REQUIRED_CONF__ = {"config": ["GOOGLE_PROJECT", "GOOGLE_DATASET"]}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bigquery_config = utils.load_config(
            ["GOOGLE_KEYFILE", "GOOGLE_PROJECT", "GOOGLE_DATASET"], self.config)
  

    def get_data(self, table, sql=None, *args, **kwargs):
        """
        A function that retrieves data from a table or using a raw SQL command.

        Args:
            table (str): name of the table to fetch data from
            sql (str): SQL command to execute

        Returns:
            pandas.DataFrame: fetched data
        """
        if sql is None:
            if table is not None:
                sql = "SELECT * FROM %s" % table
            else:
                raise Exception(
                    "Table or SQL command must be provided to BigQueryDataTransformer in order to retrieve data.")

        data = self._load_data(sql, self.bigquery_config)

        return data

    def save_data(self, data, table, *args, **kwargs):
        """
        Function to save data to a BigQuery table.

        Args:
            data (pandas.DataFrame): data to be saved
            table (str): name of the table to save the data in

        Returns: 
            None
        """
        #check if table already exists
        sql = "SELECT * FROM %s.INFORMATION_SCHEMA.TABLES" %self.bigquery_config.get("GOOGLE_DATASET")
        tables = self.get_data(None, sql=sql)

        #if table exists, see if we've added new columns or deleted columns
        table_exists = tables[tables["table_name"] == table].shape[0] > 0
        if table_exists:
           table_schema = self.get_data(table)
           new_features = [x for x in data.columns if x.upper(
           ) not in table_schema.columns.str.upper()]
           deleted_features = [
               x for x in table_schema.columns if x.upper() not in data.columns.str.upper()]

           #if we have new columns, we need to alter the table definiton
           if len(new_features) > 0:
               sql = "ALTER TABLE %s " % (table)
               for col_name in new_features:
                    col_type = data.dtypes[col_name]
                    bq_type = self.__map_pandas_col_type_to_bq_type(col_type)
                    sql = sql + ("ADD COLUMN %s %s" %
                                 (col_name.upper(), bq_type))
               self.get_data(None, sql=sql)

           #if we've deleted any features, then just add blank rows to the dataframe
           if len(deleted_features) > 0:
               for feature in deleted_features:
                   data[feature] = None

        #now we can save our data
        self._save_data(data, table, self.bigquery_config)

    def __map_pandas_col_type_to_bq_type(self, col_type):
        """
        A function that maps a pandas column type to a BigQuery column type.

        Args:
            col_type (str): pandas column type

        Returns:
            _type_: BigQuery column type
        """
        if col_type.kind == 'M':
            # datetime and timestamp columns in pandas are converted to datetime64[ns] dtype, which corresponds to 'datetime64'
            column_type = 'DATETIME'
        elif col_type.kind == 'm':
            # timedelta columns in pandas are converted to timedelta64[ns] dtype, which corresponds to 'variant'
            column_type = 'INTERVAL'
        elif col_type.kind == 'b':
            # boolean columns in pandas are converted to 'bool' dtype, which corresponds to 'boolean'
            column_type = 'BOOL'
        elif col_type.kind == 'i' or col_type.kind == 'u':
            # integer columns in pandas are converted to 'int64' dtype, which corresponds to 'integer'
            column_type = 'INT64'
        elif col_type.kind == 'f' or col_type.kind == 'c':
            # float columns in pandas are converted to 'float64' dtype, which corresponds to 'float'
            column_type = 'FLOAT64'
        else:
            # All other column types in pandas are assumed to be string columns, which correspond to 'varchar'
            column_type = 'STRING'
        return column_type

    #load data into df
    @classmethod
    def _load_data(self, sql, config, *args, **kwargs):
        """
        Function to load data from BigQuery by executing SQL query.

        Args:
            sql (str): SQL command to execute
            config (dict): Bigquery configuration information

        Returns:
            pandas.DataFrame: Fetched data from BigQuery
        """
        result = pd.DataFrame()
        client = self._get_client(config.get("GOOGLE_KEYFILE"), project=config.get("GOOGLE_PROJECT"))

        # Run a query to retrieve data from the table
        query = client.query(sql)

        # Get the results of the query as a Pandas DataFrame
        result = query.to_dataframe()

        return result

    def _save_data(self, data, table_name, config, *args, **kwargs):
        """Function to save data in form of pandas dataframe into a BigQuery table.

        Args:
            data (pandas.DataFrame): The data to save
            table_name (str): Name of table to save data
            config (dict): Configuration file for bigqury project

        """
        # Set up the BigQuery client
        client = self._get_client(config.get(
            "GOOGLE_KEYFILE"), project=config.get("GOOGLE_PROJECT"))

        # Set up the table reference
        dataset = client.dataset(config.get("GOOGLE_DATASET"))
        table = dataset.table(table_name)

        # Upload the DataFrame to BigQuery
        job_config = bigquery.LoadJobConfig()
        job_config.autodetect = True
        load_job = client.load_table_from_dataframe(data, table, job_config=job_config)

        # Wait for the job to complete
        try: 
            load_job.result()
            self.log("Successfully uploaded table %s to dataset %s" %
                     (table, config.get("GOOGLE_DATASET")))
        except Exception as e:
            self.log("Failed to upload table %s to dataset %s: %s" % 
                     (table, config.get("GOOGLE_DATASET"), str(e))) 



    def _get_client(self, key_path, project): 
        """
        Function to get a BigQuery client object.

        Args:
            key_path (str): Path to BigQuery credentials file
            project (str): Google cloud project ID

        Returns:
            google.cloud.bigquery.client - A client object for BigQuery
        """
        if key_path is not None: 
            credentials = service_account.Credentials.from_service_account_file(
                key_path, scopes=[
                    "https://www.googleapis.com/auth/bigquery"],
            )

            client = bigquery.Client(credentials=credentials,project=credentials.project_id)
        elif os.environ.get("GOOGLE_APPLICATION_CREDENTIALS") is not None: 
            client = bigquery.Client(project=project)
        else: 
            raise Exception("Must either provide GOOGLE_KEYFILE in lolpop configuration or set environment variable GOOGLE_APPLICATION_CREDENTIALS.")

        return client
    
    def _get_format(self, extension): 
        """
        Function to get the MIME content type for a given file format extension.

        Args:
            extension (str): File extension

        Returns:
            str: MIME type
        """
        if extension == ".csv": 
            return "text/csv"
        else: 
            return "application/octet-stream"


from lolpop.component.data_profiler.base_data_profiler import BaseDataProfiler
from lolpop.utils import common_utils as utils
from evidently.report import Report
from evidently.metric_preset import DataQualityPreset, DataDriftPreset
from evidently.test_suite import TestSuite
from evidently.test_preset import DataQualityTestPreset, DataDriftTestPreset, DataStabilityTestPreset
from evidently import ColumnMapping 

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class EvidentlyAIDataProfiler(BaseDataProfiler): 
    
    __REQUIRED_CONF__ = {
        "config" : ["local_dir"]
    }

    __DEFAULT_CONF__ = {
        "config": {"EVIDENTLYAI_PROFILE_REPORT_NAME": "EVIDENTLYAI_DATA_PROFILE_REPORT.HTML",
                   "EVIDENTLYAI_COMPARISON_REPORT_NAME": "EVIDENTLYAI_DATA_COMPARISON_REPORT.HTML",}
    }

    def profile_data(self, data, *args, **kwargs): 
        """Profiles data using EvidentlyAI

        Args:
            data (pd.DataFrame): A dataframe of the data to profile.  

        Returns:
            data_report (object): Python object of the report 
            file_path (string): file path of the exported report
        """
        #set up column mapping
        column_mapping = ColumnMapping()
        model_target = self._get_config("MODEL_TARGET")
        column_mapping.target = model_target
        column_mapping.prediction = "prediction"
        column_mapping.id = self._get_config("model_index")
        column_mapping.datetime = self._get_config("model_time_index")

        data_report = Report(metrics=[DataQualityPreset()])
        data_report.run(current_data=data, reference_data=None, column_mapping=column_mapping)
        file_path = "%s/%s" % (self._get_config("local_dir"), self._get_config("EVIDENTLYAI_PROFILE_REPORT_NAME"))
        data_report.save_html(file_path)
        
        return data_report, file_path

    def compare_data(self, data, prev_data, *args, **kwargs): 
        """Produces a data drift report between two data sets 
           using EvidentlyAI. 

        Args:
            data (pd.DataFrame): A dataframe of the "current" data. 
            prev_data (pd.DataFrame): A dataframe of the "historical" data. 

        Returns:
            data_report (object): Python object of the report 
            file_path (string): file path of the exported report
        """
        data_report = None
        file_path = None 
        (data, prev_data, ok) = utils.compare_data_schemas(self, data, prev_data)
        if ok: 
            column_mapping = ColumnMapping()
            column_mapping.target = check_col_exists(self._get_config("MODEL_TARGET"), data, prev_data)
            column_mapping.prediction = check_col_exists("prediction", data, prev_data)
            column_mapping.id = check_col_exists(self._get_config("model_index"), data, prev_data)
            column_mapping.datetime = check_col_exists(self._get_config("model_time_index"), data, prev_data)

            #data_comparison = TestSuite(tests=[DataStabilityTestPreset(), DataDriftTestPreset()]) #getting errors w/ the testsuite
            data_comparison = Report(metrics=[DataDriftPreset()])
            data_comparison.run(reference_data=prev_data, current_data=data, column_mapping=column_mapping)
            file_path = "%s/%s" % (self._get_config("local_dir"),
                                   self._get_config("EVIDENTLYAI_COMPARISON_REPORT_NAME"))
            data_comparison.save_html(file_path)

        return data_report, file_path
    
#drift reports error if one dataset doesn't contain something in the column mapping 
# so this ensures we don't get into a mapping that will cause an error. 
def check_col_exists(col, dfA, dfB): 
    if col in dfA.columns and col in dfB.columns: 
        return col 
    else: 
        return None 


from lolpop.component.hyperparameter_tuner.base_hyperparameter_tuner import BaseHyperparameterTuner
from lolpop.utils import common_utils as utils

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class LocalHyperparameterTuner(BaseHyperparameterTuner): 

    __REQUIRED_CONF__ = {
        "config": ["training_params", "metrics", "perf_metric"], 
        "component": ["metadata_tracker", "metrics_tracker", "resource_version_control"]
        }

    def run_experiment(self, data, model_version, *args, **kwargs): 
        """
        Runs experiments with different training parameter configurations and returns the best performing model.

        Args:
        - data: Dictionary containing input data for training and testing the model.
        - model_version: model_version object from metadata tracker

        Returns:
        - best_model: The best performing model based on hyperparameter tuning.

        Description:
        - This method generates a list of experiments by performing hyperparameter tuning with different training parameter
        configurations. For each configuration, it trains a model, saves it, makes predictions, calculates metrics,
        and logs the metrics. It then determines the best experiment based on the performance metric, saves the data splits,
        retrieves the winning experiment and model trainer data, and logs important information to the model version.
        Finally, it returns the best performing model based on the hyperparameters tuning.
        """
        # params can call different algos and user hyperparam tuning, etc 
        # first we generate a list of experiments + scores    
        exp_list = {}
        model_list = {}
        training_params = self._get_config("training_params")
        metrics = self._get_config("metrics")
        perf_metric = self._get_config("perf_metric")
        for algo in training_params: 
            grid = self._build_training_grid(training_params[algo])
            for params in grid:
                #train model 
                model, experiment = self.build_model(data, model_version, algo, params)
                #save model
                self.save_model(model, experiment)
                #make predictions
                predictions = model.transform_and_predict(data)
                #calculate metrics
                metrics_val = model.calculate_metrics(data, predictions, metrics)
                #log metrics
                self.metrics_tracker.log_metrics(experiment, metrics_val, perf_metric)
                #build experiment list
                exp_id = self.metadata_tracker.get_resource_id(experiment)
                exp_list[exp_id] = metrics_val["valid"][perf_metric]
                model_list[exp_id] = model

        #save data splits
        for k,v in data.items(): 
            vc_info = self.resource_version_control.version_data(model_version, v, key=k)
            self.metadata_tracker.register_vc_resource(model_version, vc_info, key=k, file_type="csv")

        #now, we determine overall best experiment and save into model_version
        winning_exp_id = self._get_winning_experiment(exp_list, reverse=utils.get_metric_direction(perf_metric))
        winning_exp = self.metadata_tracker.get_resource(winning_exp_id, parent=model_version, type="experiment")
        winning_exp_model_trainer = self.metadata_tracker.get_metadata(winning_exp, id="model_trainer")
        winning_exp_feature_transformer = self.metadata_tracker.get_metadata(winning_exp, id="feature_transformer_class")
        best_model = model_list.get(winning_exp_id)

        #log important stuff to model version
        self.metrics_tracker.copy_metrics(winning_exp, model_version)
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_id", data= winning_exp_id)
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_model_trainer", data= winning_exp_model_trainer)
        if winning_exp_feature_transformer is not None: 
            self.metadata_tracker.log_metadata(model_version, id="winning_experiment_feature_transformer", data=winning_exp_feature_transformer)
            winning_exp_feature_transformer_config = self.metadata_tracker.get_metadata(winning_exp, id="feature_transformer_config")
            self.metadata_tracker.log_metadata(model_version, id="winning_experiment_feature_transformer_config", data=winning_exp_feature_transformer_config)
        return best_model


from lolpop.component.base_component import BaseComponent
from typing import Any 

class BaseDataProfiler(BaseComponent): 

    def profile_data(self, data, *args, **kwargs) -> tuple[Any, str]:
        pass 

    def compare_data(self, data, prev_data, *args, **kwargs)-> tuple[Any, str]:
        pass



import typer 
import json 
from typer.core import TyperGroup 
from pathlib import Path 
from typing import List 

from lolpop.utils import common_utils as utils

class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()
    
app = typer.Typer(cls=NaturalOrderGroup, help="Test lolpop runners, pipelines, and components.")


@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))


@app.command("workflow", help="Test a workflow.")
def workflow(
    integration_class: str = typer.Argument(..., help="Integration class to test."),
    test_config: Path = typer.Argument(..., help="Location of the testing configuration file."),
    integration_config: Path = typer.Option(None, help="Location of integration configuration file. Optional. If not provided, lolpop will attempt to use the testing configurationa as the runner configuration as well."),
    integration_type: str = typer.Option("runner",
                                     help="The type of class to test: runner, pipeline, component, etc."),
    build_method: str = typer.Option("main",
                                     help="The method in the integration class to execute."),
    build_args: List[str] = typer.Option([], help="List of args to pass into build_method."),
    build_kwargs: str = typer.Option("{}", help="Dict (as a string) of kwargs to pass into build_method"),
): 
    #load class
    typer.secho("Loading integration %s" % integration_class, fg="blue")
    integration_cl = utils.load_class(
        integration_class, class_type=integration_type)
    typer.secho("Loaded %s!" % integration_class, fg="green")

    #build testing plan 
    typer.secho("Constructing test plan from configuration: %s" %test_config, fg="blue")
    test_plan, test_logger, test_recorder = utils.generate_test_plan(test_config)
    typer.secho("Test plan built! Found %s method(s) to test" %len(test_plan), fg="green")

    #intantiate class 
    if integration_config is None: 
        integration_config = test_config
    typer.secho("Initializing class %s with config file %s" %(integration_class, integration_config), fg="blue")
    integration = integration_cl(conf=integration_config)
    typer.secho("Initialized!", fg="green")

    #apply test plan to class
    typer.secho("Applying test plan to class %s" %(integration_class), fg="blue")
    integration = utils.apply_test_plan(integration, test_plan, test_recorder, test_logger)
    typer.secho("Test plan applied!", fg="green")

    if hasattr(integration, build_method):
        typer.secho("Executing %s.%s with args %s and kwargs %s" % (
            integration_class, build_method, build_args, build_kwargs), fg="blue")
        func = getattr(integration, build_method)
        func(*build_args, **json.loads(build_kwargs))
        typer.secho("Workflow completed!", fg="green")

        if test_recorder is not None: 
            typer.secho('\nPrinting Test Report: ', fg="blue")
            test_recorder.print_report()        
    else: 
        typer.secho("ERROR: Method %s not found in class %s!" %(build_method, integration_class), fg="red")


from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

from google.cloud import storage 
from google.oauth2 import service_account
from google.resumable_media.requests import ResumableUpload
from google.auth.transport.requests import AuthorizedSession
from tqdm import tqdm

import pandas as pd
import io 
import os 


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class GCSDataConnector(BaseDataConnector):
    __REQUIRED_CONF__ = {"config": ["GOOGLE_PROJECT"]}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gcs_service_account_keyfile = self._get_config("GOOGLE_KEYFILE", None)
        self.gcs_project = self._get_config("GOOGLE_PROJECT")



    def get_data(self, path, *args, **kwargs):

       """
        Load data from the given Google Cloud Storage path and return as a Pandas DataFrame.

        Parameters
        ----------
        path: str
            Path of the file to be loaded.

        Returns
        -------
        pd.DataFrame
            DataFrame with the loaded data.
        """
       df = self._load_data(path, self.gcs_service_account_keyfile, self.gcs_project)
       return df

    def save_data(self, data, path, *args, **kwargs):
        """
        Save the provided data to the given Google Cloud Storage path.

        Parameters
        ----------
        data: pd.DataFrame
            DataFrame to be saved.
        path: str
            Path where the file will be saved.

        Returns
        -------
        None
        """
        self._save_data(data, path, self.gcs_service_account_keyfile, self.gcs_project)


     #load data into df
    @classmethod
    def _load_data(self, path, keyfile, project, *args, **kwargs):
        """
        Load data from the given Google Cloud Storage path and return as a Pandas DataFrame.

        Parameters
        ----------
        path: str
            Path of the file to be loaded.
        keyfile: str
            Path to the Google Cloud Storage service account file.
        project: str
            Name of the Google Cloud Storage project.

        Returns
        -------
        pd.DataFrame
            DataFrame with the loaded data.
        """
        client = self._get_client(keyfile, project)
        bucket = path.split("/")[0]
        file_path = "/".join(path.split("/")[1:])
        extension = path.split(".")[-1]
        source_data = client.get_bucket(bucket).blob(file_path).download_as_string()

        if extension == "csv":
            data = pd.read_csv(io.StringIO(source_data), encoding="utf-8", **kwargs)
        elif extension == "parquet" or extension == "pq":
            data = pd.read_parquet(io.BytesIO(source_data), **kwargs)
        elif extension == "orc":
            data = pd.read_orc(io.BytesIO(source_data), **kwargs)
        else:
            self.log("Unsupported file type provided: %s" % extension)
        self.log("Successfully loaded data from %s into DataFrame." %path)
        
        return data 

    def _save_data(self, data, path, keyfile, project, *args, **kwargs):
        """
        Save the provided data to the given Google Cloud Storage path.

        Parameters
        ----------
        data: pd.DataFrame
            DataFrame to be saved.
        path: str
            Path where the file will be saved.
        keyfile: str
            Path to the Google Cloud Storage service account file.
        project: str
            Name of the Google Cloud Storage project.

        Returns
        -------
        None
        """
        # get client
        client = self._get_client(keyfile, project)

        #get blob based on path
        bucket = path.split("/")[0]
        file_path = "/".join(path.split("/")[1:])
        extension = path.split(".")[-1]
        blob = client.get_bucket(bucket).blob(file_path)

        format = self._get_format(extension)

        if extension == ".csv": 
            f = io.StringIO()
            data.to_csv(f, index=False)
        elif extension == ".pq" or extension == "parquet": 
            f = io.BytesIO()
            data.to_parquet(f, index=False)
        else:  
            raise Exception("Unsupported file type for path: %s" %
                            (file_path))
        f.seek(0)

        url = "https://www.googleapis.com/upload/storage/v1/b/%s/o?uploadType=resumable" % bucket
        chunksize = (1024 * 1024)  # 1MB
        upload = ResumableUpload(url, chunksize)
        transport = AuthorizedSession(credentials=client._credentials)
        upload.initiate(transport, f, {"name": blob.name}, format)

        with tqdm(total=upload.total_bytes, desc="Upload %s to %s bucket" %(file_path, bucket)) as pbar:
            while not upload.finished:
                upload.transmit_next_chunk(transport)
                pbar.update(chunksize)

        self.log("Successfully uploaded %s" %path)


    def _get_client(self, key_path, project): 
        """
        Return a Google Cloud Storage client authenticated using the provided credentials.

        Parameters
        ----------
        key_path: str
            Path to the Google Cloud Storage service account file.
        project: str
            Name of the Google Cloud Storage project.

        Returns
        -------
        storage.Client
            Authenticated GSC client object.
        """
        if key_path is not None: 
            credentials = service_account.Credentials.from_service_account_file(
                key_path, scopes=[
                    "https://www.googleapis.com/auth/cloud-platform"],
            )

            client = storage.Client(credentials=credentials,project=credentials.project_id)
        elif os.environ.get("GOOGLE_APPLICATION_CREDENTIALS") is not None: 
            client = storage.Client(project=project)
        else: 
            raise Exception("Must either provide GOOGLE_KEYFILE in lolpop configuration or set environment variable GOOGLE_APPLICATION_CREDENTIALS.")

        return client
    
    def _get_format(self, extension): 
        """
        Returns the MIME type for the provided file extension.

        Parameters
        ----------
        extension: str
            File extension.

        Returns
        -------
        str
            MIME type.
        """
        if extension == ".csv": 
            return "text/csv"
        else: 
            return "application/octet-stream"


from lolpop.component.data_profiler.base_data_profiler import BaseDataProfiler
from lolpop.utils import common_utils as utils
import sweetviz as sv

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class SweetvizDataProfiler(BaseDataProfiler): 
    
    __REQUIRED_CONF__ = {
        "config" : ["local_dir", "model_target"]
    }

    __DEFAULT_CONF__ = {
        "config": {"SWEETVIZ_PROFILE_REPORT_NAME": "SWEETVIZ_DATA_PROFILE_REPORT.HTML",
                   "SWEETVIZ_COMPARISON_REPORT_NAME": "SWEETVIZ_DATA_COMPARISON_REPORT.HTML"}
    }

    def profile_data(self, data, *args, **kwargs): 
        """Profiles data using Sweetviz

        Args:
            data (pd.DataFrame): A dataframe of the data to profile.  

        Returns:
            data_report (object): Python object of the report 
            file_path (string): file path of the exported report
        """
        model_target = self._get_config("model_target")
        feat_cfg = sv.FeatureConfig(force_num=[model_target]) #sv only supports numeric and boolean targets, so we may as well just force all categorical to be numeric
        data_report = sv.analyze(data, target_feat = model_target, feat_cfg = feat_cfg)
        file_path =  "%s/%s" %(self._get_config("local_dir"), self._get_config("SWEETVIZ_PROFILE_REPORT_NAME"))
        data_report.show_html(filepath = file_path, open_browser=False) 
        
        return data_report, file_path

    def compare_data(self, data, prev_data, *args, **kwargs): 
        """Produces a (probably data drift) report between two data sets 
           using Sweetviz. 

        Args:
            data (pd.DataFrame): A dataframe of the "current" data. 
            prev_data (pd.DataFrame): A dataframe of the "historical" data. 

        Returns:
            data_report (object): Python object of the report 
            file_path (string): file path of the exported report
        """
        model_target = self._get_config("model_target")
        feat_cfg = sv.FeatureConfig(force_num=[model_target]) 
        data_report = sv.compare([data, "Current Data"], [prev_data, "Previous Data"], target_feat = model_target, feat_cfg = feat_cfg)
        file_path =  "%s/%s" %(self._get_config("local_dir"),self._get_config("SWEETVIZ_COMPARISON_REPORT_NAME"))
        data_report.show_html(filepath = file_path, open_browser=False) 

        return data_report, file_path




import mlflow
from mlflow.tracking import MlflowClient
from functools import wraps 

def get_client(tracking_uri):
    mlflow.set_tracking_uri(tracking_uri)

    return MlflowClient(tracking_uri=mlflow.get_tracking_uri())

def get_experiment(experiment_name):

    try:
        experiment = mlflow.get_experiment_by_name(experiment_name)
        experiment_id = experiment.experiment_id
    except:
        experiment_id = mlflow.create_experiment(experiment_name)

    return experiment_id

def create_run(client, experiment_id): 
    #use mlflow.start_run instead of client.create_run because the latter doesn't set the active run 
    active_run = mlflow.active_run()
    if active_run is not None: 
        # if experiment id is the same as active run, then we are in a nested scenario
        if active_run.info.experiment_id == experiment_id: 
            return create_nested_run(active_run)
        #if experiment id is different, then we need to shut down the active run
        else: 
            stop_run()
            
    #should no longer be an active run at this point, so we can create one:
    run = mlflow.start_run(experiment_id=experiment_id)
    #run = client.create_run(experiment_id=experiment_id)
    #tag as a parent run for filtering on previous runs
    client.set_tag(run.info.run_id, "parent_run", "True")
    return run 

def create_nested_run(parent_run): 
    active_run = mlflow.active_run()
    #make sure we have the right active run
    #Note that if you start_run for an already active_run you get an error from mlflow 
    if active_run is not None: 
        #if the parent is not the active run, make it the active run so the nested run
        #is properly placed under the parent run
        if active_run.info.run_id != parent_run.info.run_id: 
            mlflow.end_run() #explicitly end run ... this used to not be necessary but seems to be now
            mlflow.start_run(run_id = parent_run.info.run_id)
    else: #we have no active run, so we need to start the parent run
        mlflow.start_run(run_id=parent_run.info.run_id)
    #create nested run under parent
    run = mlflow.start_run(experiment_id=parent_run.info.experiment_id, nested=True)
    return run 

def stop_run(): 
    mlflow.end_run()
    #if we were in a nested run then we should end the parent run too. 
    active_run = mlflow.active_run()
    if active_run is not None: 
        mlflow.end_run()

def connect(tracking_uri, experiment_name):
    client = get_client(tracking_uri)
    experiment_id = get_experiment(experiment_name)
    run = create_run(client, experiment_id) 

    return client, run

def get_run(client, run_id): 
    run = client.get_run(run_id)
    return run

#wraps logging calls around function execution
# This decorator can be used to make sure mlflow has the right active run for the object
# used. Typically, obj here will be metadata_tracker, metrics_tracker, model_respository, etc. 
# This is needed because many mlflow command live under mflow and not the mlflow client, so that when you
# are executing complex workflows, mlflow won't know what the active run is if it's being spun up in anothe process
# this was discovered when testing mlflow w/ metaflow pipelines, but it's a good thing to safeguard against in general
def check_active_mlflow_run(mlflowlib):
    def check_decorator(func, obj):
        @wraps(func)
        def wrapper(obj, *args, **kwargs):
            if mlflowlib.active_run() is None:
                run_id = obj.run.info.run_id
                uri = obj.url
                obj.log("No active mlflow run found. Reinitializing with saved values: run id %s with uri %s" %(run_id, uri))
                mlflowlib.set_tracking_uri(uri)
                mlflowlib.start_run(run_id=run_id)
            return func(obj, *args, **kwargs)
        return wrapper
    return check_decorator



from lolpop.component.base_component import BaseComponent
from lolpop.utils import common_utils as utils 

from sklearn import metrics as sk_metrics
import pandas as pd 
import numpy as np 
from typing import Any 

class BaseModelTrainer(BaseComponent): 

    __REQUIRED_CONF__ = {
        "component" : ["metadata_tracker", "resource_version_control"], 
    }

    model = None 
    mlflow_module = "you_need_to_implement_this_for_mlflow_use"
    
    #predictions = {}
    def __init__(self, params=None, *args, **kwargs): 
        #set normal config
        super().__init__(params=params, *args, **kwargs)
        #if params aren't passed in, try to retrieve from config
        if params is None or params == {}: 
            params = self._get_config("training_params", {})
        self.params = params

        #load transformer if it has one
        transformer_class = self._get_config("transformer_class")
        if transformer_class is not None: 
            transformer_config = self.get_config("transformer_config")
            transformer_cl = utils.load_class(transformer_class)
            dependent_integrations = {"component": {"logger": self.logger, "notifier": self.notifier,
                                    "metadata_tracker": self.metadata_tracker,
                                    "resource_version_control": self.resource_version_control}}
            self.feature_transformer = transformer_cl(conf=transformer_config, 
                                                      parent=self, 
                                                      is_standalone=True, 
                                                      problem_type=self.problem_type, 
                                                      dependent_integrations=dependent_integrations)
        

    def fit(self, data, *args, **kwargs) -> Any: 
        pass 

    def predict(self, data, *args, **kwargs) -> Any:
        pass

    def predict_df(self, df, *args, **kwargs) -> Any:
        pass

    def predict_proba_df(self, df, *args, **kwargs) -> Any:
        pass

    def get_artifacts(self, id, *args, **kwargs) -> dict[str, Any]:
        pass

    def transform_and_fit(self, data_dict, *args, **kwargs) -> Any: 
        #fit feature transformer
        self.fit_data(data_dict["X_train"], data_dict["y_train"])

        #transform features 
        transformed_data_dict = self._transform_dict(data_dict)

        #finally, fit model
        return self.fit(transformed_data_dict, *args, **kwargs)
    
    def transform_and_predict(self, data, *args, **kwargs) -> Any: 
        #transform data 
        transformed_data = self._transform_dict(data)
        
        #finally, call predict on transformed data         
        return self.predict(transformed_data, *args, **kwargs)
    
    def transform_and_predict_df(self, data, *args, **kwargs) -> Any: 
        #transform dataframe
        transformed_data = self.transform_data(data)

        #get predictions
        return self.predict_df(transformed_data, *args, **kwargs)

    def transform_and_predict_proba_df(self, data, *args, **kwargs) -> Any: 
        #transform dataframe
        transformed_data = self.transform_data(data)

        #get prediction probabilities
        return self.predict_proba_df(transformed_data, *args, **kwargs)

    def fit_transform_data(self, X_data, y_data, *args, **kwargs) -> Any:
        #fit data 
        self.fit_data(X_data, y_data)

        #transform data 
        data_out = self.transform_data(X_data)
        return data_out
        
    def fit_data(self, X_data, y_data=None, *args, **kwargs) -> Any:

        if hasattr(self, "feature_transformer"):
            self.log(
                "Fitting feature transformer on data for trainer %s." % self.name)
            self.feature_transformer.fit(X_data, y_data)
            return self.feature_transformer
        else:
            self.log("No feature transformer found for trainer %s." %
                     self.name)
            return None 

    def _transform_dict(self, data, *args, **kwargs) -> dict[Any]: 
        data_dict = data.copy() 

        #data is a dict of train/vali/test so we need to transform all X_ datasets
        for key in data_dict.keys():
            if key.startswith("X_"):
                data_dict[key] = self.transform_data(data_dict[key])

        return data_dict

    def transform_data(self, data, *args, **kwargs) -> Any: 
        data_out = data.copy() 

        if hasattr(self, "feature_transformer"):
            self.log("Running feature transformer on data for trainer %s." %self.name)
            data_out = self.feature_transformer.transform(data_out)
        else: 
            self.log("No feature transformer found for trainer %s." %self.name)

        return data_out


    def save(self, experiment, *args, **kwargs): 
        """Save the model.

        Args:
            experiment (object): experiment to save model into
        """
        algo = type(self).__name__
        vc_info = self.resource_version_control.version_model(experiment, self.model, algo=algo)
        experiment_metadata = {
            "training_params" : self.params,
            "model_trainer" : algo
        }
        self.metadata_tracker.register_vc_resource(experiment, vc_info, additional_metadata = experiment_metadata)

        if hasattr(self, "feature_transformer"): 
            self.feature_transformer.save(experiment)

        experiment_id = self.metadata_tracker.get_resource_id(experiment, type="experiment")
        model_artifacts = self.get_artifacts(experiment_id) or {}
        for k, v in model_artifacts.items():
            self.metadata_tracker.log_artifact(experiment, k, v)

    def load(self, *args, **kwargs) -> Any: 
        #load feature transformer
        pass 

    def _get_model(self, *args, **kwargs) -> Any:
        return self.model 
    
    #used to set model object. 
    def _set_model(self, model, *args, **kwargs): 
        self.model = model 
    #def _get_predictions(self): 
    #    return self.predictions

    def _get_transformer(self, *args, **kwargs) -> Any: 
        return self.feature_transformer
    
    def _set_transformer(self, transformer, *args, **kwargs): 
        self.feature_transformer = transformer

    def calculate_metrics(self, data, predictions, metrics, *args, **kwargs) -> dict[str, float]: 
        """Calculates metrics on the trained model. 

        Args:
            data (dict): dictonary of train/test/validation data 
            predictions (object): 
            metrics (list): list of metrics to calculate

        Returns:
            metrics_out: dictionary of computed metrics
        """
        metrics_out = {"train" : {}, "valid" : {}, "test" : {}}

        test_exists = data.get("y_test") is not None
        valid_exists = data.get("y_valid") is not None

        if self.problem_type == "classification": 
            multi_class = False
            average="binary"
            num_classes = len(data["y_train"].unique())
            if num_classes > 2: 
                multi_class = True 
                average = "weighted"

        for metric in metrics: 
            if metric == "accuracy": 
                metrics_out["train"][metric] = sk_metrics.accuracy_score(data["y_train"], predictions["train"])
                if valid_exists: 
                    metrics_out["valid"][metric] = sk_metrics.accuracy_score(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.accuracy_score(data["y_test"], predictions["test"])
            
            elif metric == "f1": 
                metrics_out["train"][metric] = sk_metrics.f1_score(data["y_train"], predictions["train"], average = average)
                if valid_exists: 
                    metrics_out["valid"][metric] = sk_metrics.f1_score(data["y_valid"], predictions["valid"], average = average)
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.f1_score(data["y_test"], predictions["test"], average = average)
            
            elif metric == "rocauc": 
                metrics_out["train"][metric] = sk_metrics.roc_auc_score(data["y_train"], predictions["train_proba"], average = average, multi_class = "ovr")
                if valid_exists: 
                    metrics_out["valid"][metric] = sk_metrics.roc_auc_score(data["y_valid"], predictions["valid_proba"], average = average, multi_class = "ovr")
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.roc_auc_score(data["y_test"], predictions["test_proba"], average = average, multi_class = "ovr")
            
            elif metric == "prauc": 
                if not multi_class: #only works for binary
                    metrics_out["train"][metric] = sk_metrics.average_precision_score(data["y_train"], predictions["train_proba"][:,1])
                    if valid_exists: 
                        metrics_out["valid"][metric] = sk_metrics.average_precision_score(data["y_valid"], predictions["valid_proba"][:,1])
                    if test_exists:
                        metrics_out["test"][metric] = sk_metrics.average_precision_score(data["y_test"], predictions["test_proba"][:,1])
            
            elif metric == "precision": 
                metrics_out["train"][metric] = sk_metrics.precision_score(data["y_train"], predictions["train"], average = average)
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.precision_score(data["y_valid"], predictions["valid"], average = average)
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.precision_score(data["y_test"], predictions["test"], average = average)
           
            elif metric == "recall": 
                metrics_out["train"][metric] = sk_metrics.recall_score(data["y_train"], predictions["train"], average = average)
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.recall_score(data["y_valid"], predictions["valid"], average = average)
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.recall_score(data["y_test"], predictions["test"], average = average)
            
            elif metric == "mse": 
                metrics_out["train"][metric] = sk_metrics.mean_squared_error(data["y_train"], predictions["train"])
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.mean_squared_error(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.mean_squared_error(data["y_test"], predictions["test"])
            
            elif metric == "rmse": 
                metrics_out["train"][metric] = sk_metrics.mean_squared_error(data["y_train"], predictions["train"], squared=False)
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.mean_squared_error(data["y_valid"], predictions["valid"], squared=False)
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.mean_squared_error(data["y_test"], predictions["test"], squared=False)
            
            elif metric == "mae": 
                metrics_out["train"][metric] = sk_metrics.mean_absolute_error(data["y_train"], predictions["train"])
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.mean_absolute_error(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.mean_absolute_error(data["y_test"], predictions["test"]) 
                
            elif metric == "mape": 
                metrics_out["train"][metric] = sk_metrics.mean_absolute_percentage_error(data["y_train"], predictions["train"])
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.mean_absolute_percentage_error(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.mean_absolute_percentage_error(data["y_test"], predictions["test"])

            elif metric == "mdae": 
                metrics_out["train"][metric] = sk_metrics.median_absolute_error(data["y_train"], predictions["train"])
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.median_absolute_error(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.median_absolute_error(data["y_test"], predictions["test"])

            elif metric == "smape":
                metrics_out["train"][metric] = symmetric_mean_absolute_percentage_error(data["y_train"], predictions["train"])
                if valid_exists:
                    metrics_out["valid"][metric] = symmetric_mean_absolute_percentage_error(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = symmetric_mean_absolute_percentage_error(data["y_test"], predictions["test"])
                
            elif metric == "r2":  
                metrics_out["train"][metric] = sk_metrics.r2_score(data["y_train"], predictions["train"])
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.r2_score(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.r2_score(data["y_test"], predictions["test"])

            elif metric == "msle":
                metrics_out["train"][metric] = sk_metrics.mean_squared_log_error(data["y_train"], predictions["train"])
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.mean_squared_log_error(data["y_valid"], predictions["valid"])
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.mean_squared_log_error(data["y_test"], predictions["test"])
            
            elif metric == "rmsle":
                #predictions can be negative, which will break this calculation.
                #if you're using rmsle we'll assume you intended non-negative predictions
                for key in set(["train", "valid", "test"]).intersection(predictions.keys()): 
                    predictions[key] = [max(x,0) for x in predictions[key]]
                metrics_out["train"][metric] = sk_metrics.mean_squared_log_error(data["y_train"], predictions["train"], squared=False)
                if valid_exists:
                    metrics_out["valid"][metric] = sk_metrics.mean_squared_log_error(data["y_valid"], predictions["valid"], squared=False)
                if test_exists:
                    metrics_out["test"][metric] = sk_metrics.mean_squared_log_error(data["y_test"], predictions["test"], squared=False)
  
            

        return metrics_out

    def build_model(self, data, model_version, *args, **kwargs) -> tuple[Any, Any]: 
        """Trains a single model

        Args:
            data (dict): data dictionary of train/test/validation data
            model_version (object): model version

        Returns:
            model: the trained model
            exp: experiment where the model was trained
        """
        #fit model
        model_obj = self.transform_and_fit(data)

        #create experiment and save model 
        experiment = self.metadata_tracker.create_resource(id=None, type="experiment", parent=model_version)
        self.save(experiment)

        #get predictions
        predictions = self.transform_and_predict(data)

        #calculate metrics
        metrics_val = self.calculate_metrics(data, predictions, self._get_config("metrics"))

        #log stuff
        self.metrics_tracker.log_metrics(experiment, metrics_val, self._get_config("perf_metric"))
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_id", data=self.metadata_tracker.get_resource_id(experiment))
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_model_trainer", data = type(self).__name__)

        #save splits
        for k,v in data.items(): 
            vc_info = self.resource_version_control.version_data(model_version, v, key=k)
            self.metadata_tracker.register_vc_resource(model_version, vc_info, key=k, file_type="csv")

        return self, experiment 

    def rebuild_model(self, data, model_version, *args, **kwargs) -> tuple[Any, Any]:
        """Trains model using all available data in the data dictionary

        Args:
            data (dict): dictionary of train/test/validation data
            model_version (object): model version which we want to retrain

        Returns:
            model: newly trained model
            exp: experiment in which the model was trained
        """

        has_test = ("X_test" in data.keys())

        #merge all dfs 
        df_X = pd.concat([data["X_train"], data["X_valid"]])
        df_y = pd.concat([data["y_train"], data["y_valid"]])
        if has_test: 
            df_X = pd.concat([df_X, data["X_test"]])
            df_y = pd.concat([df_y, data["y_test"]])
        train_data = {"X_train" : df_X, "y_train": df_y}

        #get params, class from winning experiment 
        #winning_experiment = self.metadata_tracker.get_winning_experiment(model_version)
        #params = self.metadata_tracker.get_metadata(winning_experiment, "training_params")
        #model_trainer = self.metadata_tracker.get_metadata(winning_experiment, "model_trainer")
        
        #rebuild model on all data 
        model, exp = self.build_model(train_data, model_version)                        
        
        return model, exp
    

def symmetric_mean_absolute_percentage_error(y_true, y_pred):
    """computes symmetric mean absolute percentage error

    Args:
        y_true : actuals
        y_pred : predictions

    Returns:
        float: symmetric mean absoulte percentage error
    """
    epsilon = np.finfo(np.float64).eps
    smape = (2 * np.abs(y_pred - y_true) / np.maximum(np.abs(y_pred) + np.abs(y_true), epsilon))
    return np.average(smape[~np.isnan(smape)])



import subprocess 
import asyncio
import os 
import sys
import json 
from importlib import import_module, util as import_util 
from git import Repo
from omegaconf import OmegaConf, dictconfig, listconfig
from functools import wraps 
import pandas as pd 
from pathlib import Path 
import time
from inspect import getsource
import types
from inspect import isclass

async def async_cmd(cmd): 
    p = await asyncio.create_subprocess_exec(cmd, universal_newlines=True, shell=False)
    return p 

#shell command wrapper
def execute_cmd(cmd, logger=None, run_async=False, run_background=False):
    if logger: 
        logger.log("Executing command: `%s`" %(" ".join(cmd)))      
    if run_async: 
        p = async_cmd(cmd)
        return (p,0)
    elif run_background: 
        p = subprocess.Popen(cmd, universal_newlines=True, shell=False, close_fds=True)
        return (p, 0)
    else: 
        p = subprocess.run(cmd, stdout=subprocess.PIPE,
                       universal_newlines=True, shell=False)
        return (p.stdout, p.returncode)
    
# pulls keys out of config or gets from env variable
# i.e. either handle secret manager before getting here and pass in 
# or populate via env vars
def load_config(config_keys, conf): 
    config = {}
    for key in config_keys: 
        if key in conf.keys():
            config[key]=conf[key]
        elif key.lower() in conf.keys(): #handle cases where we're looking for env var but it's overridden in conf w/ lowercase 
            config[key] = conf[key.lower()]
        else: 
            config[key] = os.getenv(key)
    return config

#comits a single file to git
def git_commit_file(file_path, repo_path=None, msg="Commiting file from lolpop", push=False, path_from_root=None, logger=None):
    repo = Repo(repo_path,search_parent_directories=True)
    if path_from_root is not None: 
        file_path = "%s/%s" %(path_from_root, file_path)
    repo.index.add(file_path)
    hexsha = repo.index.commit(msg).hexsha

    if logger:
        logger.log("Committed file %s." % (file_path))

    if push: 
        get_push_repo(repo_path, logger)

    return hexsha

def get_push_repo(repo_path=None, logger=None): 
    repo = Repo(repo_path, search_parent_directories=True)
    origin = repo.remotes[0]
    result = origin.push()
    if logger:
        logger.log("Pushed repository %s. Result: %s" % (repo.working_dir, result))
    return result

def load_plugins(plugin_paths=None):
    plugins = []
    if plugin_paths is not None and len(plugin_paths) > 0:
        for dir in plugin_paths:
            if dir.exists():
                plugin = load_plugin(dir)
                plugins.append(plugin)
    return plugins

def load_plugin(plugin_path, obj=None): 
    #get plugin directory path and name
    plugin_path = plugin_path.resolve() #in case absolute path is not passed in in yaml
    if plugin_path.is_dir(): 
        plugin_dir = str(plugin_path)
        plugin_name = plugin_path.name
    elif plugin_path.is_file():
        plugin_dir = str(plugin_path.parent)
        plugin_name = plugin_path.stem
    else: 
        raise Exception("Invalid plugin path: %s. Path is not a file nor a directory." %plugin_path)
    #append directory to sys path
    if plugin_dir: 
        sys.path.append(plugin_dir)
    #import module
    mod = None
    #there's only a module to load if we're passed a file 
    if plugin_path.is_file(): 
        if plugin_name: 
            mod = import_module(plugin_name)
    #if object passed in, append plugin dir to plugin_paths config
    #this is mainly used for local data transformer component
    if obj: 
        plugin_paths = obj._get_config("plugin_paths",[]) 
        plugin_paths.append(plugin_dir)
        obj._set_config("plugin_paths", plugin_paths)
    return mod

#load class object
def load_class_obj(class_name, class_type="component", parent="lolpop"): 
    try: 
        module = import_module("%s.%s" %(parent, class_type))
        cl = getattr(module, class_name)
    except: 
        module = import_module(parent)
        cl = getattr(module, class_name)
    return cl

def load_class(class_name, class_type="component", plugin_mods = [], self_obj=None): 
    try:
        cl = load_class_obj(class_name, class_type)
    except: 
        try: 
            if self_obj: 
                self_obj.log("Unable to find class %s in build-in resources. Searching extensions..." %class_name)
            cl = load_class_obj(class_name, class_type="extension")
            if self_obj:
                self_obj.log("Found class %s in extensions!" %class_name)
        except: 
            if self_obj:
                self_obj.log(
                    "Unable to find class %s in extensions. Searching plugins modules in %s..." %(class_name, str(plugin_mods)))
            cl = load_class_from_plugin(class_name, plugin_mods, class_type)
            if cl is not None: 
                if self_obj: 
                    self_obj.log("Found class %s in plugins!" % class_name)
            else: 
                if self_obj:
                    self_obj.log("Unable to find class %s in plugins!" %class_name)
    return cl 

#load class object from plugins
def load_class_from_plugin(class_name, plugin_mods, class_type="component"):
    #try to load from each plugin
    cl = None 
    for plugin in plugin_mods: 
        try: 
            cl = load_class_obj(class_name, class_type=class_type, parent=plugin)
            break 
        except Exception as e: 
            #if multiple plugins are used, the one we are looking for will not be in most of them, so just ignore any errors
            continue 
    return cl

#loads a module from a file
def load_module_from_file(file_path): 
    spec = import_util.spec_from_file_location(file_path.stem, file_path)
    mod = import_util.module_from_spec(spec)
    sys.modules[spec.name]=mod #need to do this to properly register module
    spec.loader.exec_module(mod)
    return mod 

def register_integration_class(self_obj, conf, integration_type_type, 
                               integration_type="component",
                               integration_framework=None,
                               default_class_name=None,
                               problem_type=None,
                               dependent_integrations=None,
                               plugin_mods=None,
                               decorators=None,
                               *args, **kwargs): 
    if dependent_integrations is None: 
        dependent_integrations = {}
    if plugin_mods is None: 
        plugin_mods = []
    if decorators is None: 
        decorators = []
    obj = None 
    integration_class_name = conf.get(integration_type, {}).get(integration_type_type, default_class_name)
    if integration_class_name is not None: 
        cl = load_class(integration_class_name, 
                        class_type=integration_type,
                        plugin_mods=plugin_mods, 
                        self_obj=self_obj)
        if cl is not None: 
            if len(decorators) > 0: 
                cl = apply_decorators(cl, decorators, integration_type=integration_type)
            if integration_framework is None and self_obj.integration_framework is not None: #try to find integration framework from children
                arr_children = [x for x in self_obj.integration_framework.children if x.id == integration_type]
                if len(arr_children)>0: 
                    integration_framework= arr_children[0] #should only be one child w/ that id
            obj = cl(conf = conf.get(integration_type_type),
                     parent=self_obj, 
                     integration_framework=integration_framework,
                     problem_type=problem_type,
                     dependent_integrations=dependent_integrations,
                     plugin_mods=plugin_mods,
                     decorators=decorators, 
                     *args, **kwargs)
            setattr(self_obj, integration_type_type, obj)
    return obj

def lower_conf(conf): 
    return {k.lower():v for k,v in conf.items()}
    
#Need to revisit. Seems like this is not actually merging on update, it's reassigning. 
#As a result, some default conf can get lost
def copy_config_into(conf, default_conf): 
    #have to use omegaconf so types match
    default_copy = OmegaConf.create(default_conf.copy())
    updated_conf = OmegaConf.merge(default_copy, conf)
    return updated_conf

#validates configuration with the required conf specified
def validate_conf(config, required_conf):
    conf = config.copy()

    missing = {}
    total_missing = 0
    if required_conf is not None: 
        if conf is None: 
            conf = {}
        #map everything to lowercase keys to avoid capitalization mismatch
        lconf = lower_conf(conf)
        lrequired_conf = lower_conf(required_conf)
        for k,v in lrequired_conf.items(): 
            missing_k = []
            for val in v: 
                #value can be of form "key|values" which specifies the value(s) allowed
                #this is mainly used to specify the classes supported
                #i.e. something like "MetadataTracker|MLFlowMetadataTracker"
                if "|" in val: 
                    val_arr = val.split("|")
                    val_key = val_arr[0]
                    val_value = val_arr[1].split(",")
                else: 
                    val_key = val
                    val_value = None
                #check if key,value exist in config
                if lconf.get(k,{}).get(val_key.lower(),None) is None: 
                    #if not provided in config, check environment variables 
                    if os.getenv(val_key.upper(),None) is None: 
                        missing_k.append(val) 
                elif val_value is not None: 
                    if lconf.get(k,{}).get(val_key.lower()) not in val_value: 
                        missing_k.append(val)
            missing[k] = missing_k
        for v in missing.values(): 
            total_missing = total_missing + len(v)
    return missing, total_missing

#replaces variable values with their actual values. 
#conf allows pointing to other variable via $node1.node2.node3, etc...
def resolve_conf_variables(conf, main_conf=None):
    if main_conf is None: #need to always have the full conf lineage because variables can pont to arbitrary nodes
        main_conf = conf
    if conf is not None and len(conf) > 0: 
        for k,v in conf.items(): 
            if isinstance(v, dictconfig.DictConfig):
                OmegaConf.update(conf, k, resolve_conf_variables(v, main_conf)) 
            elif isinstance(v, str): 
                if v[0] == "$": 
                    resolved_value = get_conf_value(v[1:], main_conf)
                    if resolved_value: 
                        OmegaConf.update(conf,k, resolved_value)
                    else: 
                        raise Exception("Unable to resolve configuration value: %s" %v)
    return conf

# returns configuration from conf object (file location or python dict)
def get_conf(conf_obj): 
    if isinstance(conf_obj, str) or isinstance(conf_obj, Path):
        conf = OmegaConf.load(conf_obj)
    elif isinstance(conf_obj, dict): 
        conf = OmegaConf.create(conf_obj)
    elif isinstance(conf_obj, dictconfig.DictConfig):
        conf = conf_obj
    elif conf_obj is None: 
        conf = {}
    else: 
        raise Exception("Invalid configuration. Configuration must be a file or a a dict.")
    
    return conf 

#returns node value in conf specified by var in the form 'node1.node2.node3...'
def get_conf_value(var, conf): 
    var_arr = var.split(".")
    out = conf
    try: 
        for i in var_arr: 
            out = out.get(i)
    except: 
        out = None 
    return out 

def get_plugin_mods(self_obj, plugin_paths=None, file_path=None):
    # if no plugin_dir is provided, then try to use the parent directory.
    # if the parent directory is lolpop, then it is a built-in runner and we can ignore
    if plugin_paths is None: 
        plugin_paths = []
    if file_path is not None:
        #directory should be something like <module_name>/<runner>/<runner_type>/<runner_class>.py
        plugin_dir = os.path.dirname(os.path.dirname(
            os.path.dirname(os.path.realpath(file_path))))
        plugin_paths.append(plugin_dir)

    plugin_paths = [Path(dir) for dir in plugin_paths if dir != "lolpop"]

    #load up all plugin modules
    plugins = load_plugins(plugin_paths)
    plugin_mods = []
    for plugin in plugins: 
        if plugin is not None: 
            plugin_mods.append(plugin.__name__)
    
    return plugin_mods

#for a given lolpop object, get all plugin paths in children
def get_all_plugin_paths(obj):
    children = [x for x in dir(obj) if not x.startswith("_")]
    plugin_paths=[]

    for child in children: 
        child_obj = getattr(obj, child)
        if hasattr(child_obj, "_get_config"): 
            paths = child_obj._get_config("plugin_paths", None)
            if paths is not None: 
                plugin_paths = plugin_paths + paths
    
    plugin_paths = [Path(path) for path in set(plugin_paths)]

    return plugin_paths

#wraps logging calls around function execution
def log_execution(level="DEBUG", start_msg = None, end_msg = None, timeit=True):
    def log_decorator(func, obj):
        @wraps(func)
        def wrapper(obj, *args, **kwargs):
            start = start_msg
            end = end_msg
            if not start: 
                start = "Starting execution of %s.%s" %(obj.__module__, func.__name__)
            if not end: 
                end = "Finished execution of %s.%s" %(obj.__module__, func.__name__)
            obj.log(start, level)
            obj.log("args: %s, kwargs: %s" %(args, kwargs), "TRACE")
            start_time = time.process_time()
            result = func(obj, *args, **kwargs)
            end_time = time.process_time()
            if timeit: 
                end = end + ". Completed in %s seconds." %(str(end_time - start_time))
            obj.log(end, level)
            return result
        return wrapper
    return log_decorator

#wraps function around error handling
def error_handler(func, obj):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            #prevent unnecessary nesting
            if str(e).startswith("An error occurred"): 
                raise e
            else: 
                raise Exception(f"An error occurred while executing {obj.__module__}.{func.__name__}: {e}")
    return wrapper

# decorate all public methods in a class with decorator
# should only be applied to class definition
# decorators can be a single decorator or list of decorators that will be applied in reverse order
# i.e. the first decorator in the list would be the topmost decorator if listed out
# (this seems like the intuitive thing to do)
# Update: modified to also support being applied to a class instance.
# Reason being that we can apply this to a runner instance. This allows us to dynamically 
# decorate a runner, even though the decoration doesn't happen until after object instantiation
def decorate_all_methods(decorators, skippable_methods=["log", "notify"]):
    def decorate(obj):
        decorator_list = decorators
        if not isinstance(decorator_list, list):
            decorator_list=[decorator_list]
        for attr in dir(obj):
            # we can decorate a class or an instance (in the case of a runner, which will already
            # have been instantiated when we try to decorate it). In the latter case we need to 
            # handle it a little differently
            cls=obj
            if not isclass(cls): 
                cls=type(obj)
            if callable(getattr(obj, attr)) and not attr.startswith("_") and attr not in skippable_methods:
                func = getattr(cls,attr)
                for decorator in decorator_list[::-1]: 
                    func = decorator(func, cls)
                if isclass(obj): 
                    setattr(obj, attr, func)
                else: #if it's an instance then we need to turn the function back into a bound method
                    setattr(obj, attr, types.MethodType(func, obj))
        return obj
    return decorate

#sets up decorators in provided configuration
def set_up_decorators(obj, conf, 
                      plugin_mods=None, bind_decorators=True, 
                      decorator_integration_type="decorator", 
                      dependent_integrations=None): 
    if dependent_integrations is None: 
        dependent_integrations = {}
    #set up all decorator classes 
    decorators = []
    decorators_conf = conf.get(decorator_integration_type, {})
    for decorator_type, decorator in decorators_conf.items(): 
        decorator_cl = load_class(decorator, plugin_mods=plugin_mods,self_obj=obj)
        decorator_conf = conf.get(decorator_type,{"config":{}})
        decorator_obj = decorator_cl(
            conf=decorator_conf,
            is_standalone=True,
            dependent_integrations=dependent_integrations)
        if bind_decorators: 
            setattr(obj, decorator_type, decorator_obj)
        decorators.append(decorator_obj)

    return decorators

#apply decorators to the provided class 
def apply_decorators(cls, decorators, integration_type="component"): 
    #apply decorators
    decorator_list = []
    for decorator in decorators:
        integration_types = decorator._get_config("integration_types", ["component"])
        integration_classes = decorator._get_config("integration_classes", [])
        #apply a decorator if cls is in integration_classes or it's part of the integration_types
        if integration_classes is not None and len(integration_classes) > 0:
            apply_decorator = cls.__name__ in integration_classes
        else: 
            apply_decorator = integration_type in integration_types
        if apply_decorator:
            decorator_method = getattr(decorator,decorator._get_config("decorator_method"))
            decorator_list.append(decorator_method)
    cls = decorate_all_methods(decorator_list)(cls)
    
    return cls

#def wrap_runner_functions(obj, decorators):
#        decorator_list = decorators
#        if not isinstance(decorator_list, list):
#            decorator_list=[decorator_list]
#        for attr in dir(obj):
#            if callable(getattr(obj, attr)) and not attr.startswith("_") and attr !="log" and attr !="notify":
#                func = getattr(type(obj),attr)
#                for decorator in decorator_list[::-1]: 
#                    decorator_method = getattr(decorator, decorator._get_config("decorator_method"))
#                    func = decorator_method(func, type(obj))
#                setattr(obj, attr, types.MethodType(func, obj))
#        return obj
    

#compares two dataframes and tries to convert columns so that match types if they don't otherwise
def compare_data_schemas(obj, data, prev_data):
    #check to see if schemas are equal
    ok = True 
    if not data.equals(prev_data): 
        #if they are not equal, try to set schemas to be the same. 
        ok = False
        try: 
            data = convert_col_types(data, prev_data)
            ok = True
        except Exception as e: 
            obj.notify("Data Comparison failed. New and old dataframes do not have compatible column types. Error: %s" %str(e), "ERROR")
            raise 
    return data, prev_data, ok

def convert_col_types(data, prev_data):
    df_a = data.copy()
    df_b = prev_data.copy()
    for col in df_a.columns:
        if col in df_b.columns: #if col is net new, we can skip
            if df_a[col].dtype != df_b[col].dtype:
                df_a[col] = df_a[col].astype(df_b[col].dtype)
    return df_a

#True = higher is better (descending)
def get_metric_direction(perf_metric): 
    reverse = True 
    if perf_metric in ["mse", "rmse", "mae", "mdae", "mape", "smape", "rmsle", "msle"]: 
        reverse = False 
    return reverse

#labels should be something like data["y_train"].unique()
def get_multiclass(labels): 
    classification_type = "binary"
    num_classes = len(labels)
    if num_classes > 2: 
        classification_type = "multiclass"

    return classification_type


def create_df_from_file(source_file, engine="pyarrow", **kwargs): 
    _, source_file_type = str(source_file).split("/")[-1].split(".")
    data = pd.DataFrame()
    if source_file_type == "csv":
        data = pd.read_csv(source_file, engine=engine, **kwargs)
    elif source_file_type in ["parquet", "pq"]:
        data = pd.read_parquet(source_file, engine=engine, **kwargs)
    else:
        raise Exception("Unsupported file type: %s" % source_file_type)
    
    return data

def chunker(seq, size):
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


#generates a test plan from a configuration file
def generate_test_plan(config, test_plan=None, parent=None, logger=None, test_recorder=None): 
    if test_plan is None: 
        test_plan = {} 

    config = get_conf(config)
    components = config.pop("component", {})
    #set up logger
    if logger is None: 
        logger_cl = load_class(components.get("test_logger",None))
        if logger_cl is not None: 
            logger_conf = config.pop("test_logger",{})
            logger = logger_cl(conf=logger_conf)

    #set up test_recorder
    if test_recorder is None:
        test_recorder_cl = load_class(components.get("test_recorder", "LocalTestRecorder"))
        if test_recorder_cl is not None:
            test_recorder_conf = config.pop("test_recorder", {})
            test_recorder = test_recorder_cl(conf=test_recorder_conf)

    #extract tests
    test_plan = extract_tests(config, test_plan, parent)
    
    return test_plan, logger, test_recorder


def extract_tests(config, test_plan=None, parent=None):
    if test_plan is None: 
        test_plan = {}
    config = get_conf(config)

    #extract tests
    test_list = config.pop("tests", {})
    for method in test_list:
        key = method
        if parent is not None:
            key = "%s.%s" % (parent, key)
        test_plan[key] = test_list[method]

    #anything else should be dependents + tests
    for resource in config.keys():
        key = resource
        if parent is not None:
            key = "%s.%s" % (parent, key)
        test_plan = extract_tests(config[resource], test_plan=test_plan,
                                          parent=key)
        
    return test_plan

#apply test plan to the given object
#this replaces the specified function with a wrapper that now includes 
# the pre- and post-hooks defined in the testing config
def apply_test_plan(runner, test_plan, test_recorder, test_logger=None): 

    # if no specific test logger is provided, just use the runner's log method
    if test_logger is None: 
        test_logger = runner

    for method,tests in test_plan.items(): 
        fn, obj = get_method(runner, method)
        prehooks = load_hooks(tests.get("pre-hooks"))
        posthooks = load_hooks(tests.get("post-hooks"))

        if len(prehooks) > 0 or len(posthooks) > 0: 
            runner = set_method(runner, method, test_plan_decorator(
                obj, test_logger, test_recorder, prehooks, posthooks)(fn))
            
    return runner

#load hooks from file. 
# Note: the "test" function in the hook file is the entry point
def load_hooks(hook_list): 
    hook_list_out = []
    if hook_list is not None: 
        hook_list_out = [
            (Path(x.get("hook_file")).stem, 
             getattr(load_plugin(Path(x.get("hook_file"))), x.get("hook_method","test")),
             x.get("hook_on_error", "continue").lower()) 
             for x in hook_list
        ]
    return hook_list_out

#retrieves the method on a runner path. Here 'method' is dot notation i.e. "runner.pipeline.component.method"
def get_method(obj, method):
    method_arr = method.split(".")
    out = obj, None
    try: 
        for i in method_arr:
            parent = obj
            obj = getattr(obj, i)
        out = obj, parent
    except: 
        out = None, None 
    return out 

#sets the method for the object to the specified function
def set_method(obj, method, fn):
    method_arr = method.split(".")

    if len(method_arr) == 1: 
        setattr(obj, method_arr[0], fn)
    else: 
        setattr(obj, method_arr[0], 
                set_method(getattr(obj, method_arr[0]),".".join(method_arr[1:]),fn))
        
    return obj 

#decorator used with a test plan to apply pre- and post-hooks to a function 
def test_plan_decorator(obj, logger, recorder, prehooks, posthooks, level="DEBUG"):
    def test_decorator(func): 
        @wraps(func)
        def wrapper(*args, **kwargs):
            for prehook in prehooks:
                obj.log("Running pre-hook %s.%s" %(prehook[0], prehook[1].__name__), level)
                try: 
                    passed = prehook[1](obj, *args, **kwargs)
                    msg_out = None
                    if isinstance(passed, tuple):
                        msg_out = passed[1]
                        passed = passed[0]
                    recorder.record_test(obj, func, prehook[0], prehook[1], passed, msg_out)
                    if passed: 
                        logger.log("Pre-hook %s.%s passed." %
                                   (prehook[0], prehook[1].__name__), "INFO")
                    else: 
                        logger.log("Pre-hook %s.%s failed" %
                                   (prehook[0], prehook[1].__name__), "WARN")
                    if msg_out is not None: 
                        logger.log("Pre-hook message: %s" %msg_out)
                    obj.log("Pre-hook %s.%s finished." %(prehook[0], prehook[1].__name__), level)
                except Exception as e: 
                    logger.log("Error occured when running pre-hook %s.%s: %s" % (prehook[0], prehook[1].__name__, str(e)))
                
                
            result = func(*args, **kwargs)
            
            for posthook in posthooks:
                obj.log("Running post-hook %s.%s" %
                        (posthook[0], posthook[1].__name__), level)
                try: 
                    passed = posthook[1](obj, result, *args, **kwargs)
                    msg_out = None
                    if isinstance(passed, tuple): 
                        msg_out = passed[1] 
                        passed = passed[0]
                    recorder.record_test(obj, func, posthook[0], posthook[1], passed, msg_out)
                    if passed:
                        logger.log("Post-hook %s.%s passed." %
                                   (posthook[0], posthook[1].__name__), "INFO")
                    else:
                        logger.log("Post-hook %s.%s failed" %
                                   (posthook[0], posthook[1].__name__), "WARN")
                    if msg_out is not None:
                        logger.log("Pre-hook message: %s" % msg_out)
                    obj.log("Post-hook %s.%s finished." %
                            (posthook[0], posthook[1].__name__), level)
                except Exception as e:
                    logger.log(
                        "Error occured when running post-hook %s.%s: %s" % (posthook[0], posthook[1].__name__, str(e)))
            return result
        return wrapper
    return test_decorator


def set_up_default_integrations(obj, conf, 
                                default_integrations,
                                plugin_mods=None,
                                skip_config_validation=False,
                                dependent_integrations=None, 
                                decorators=None
                              ):
    if dependent_integrations is None: 
        dependent_integrations = {}
    if plugin_mods is None: 
        plugin_mods = []
    if decorators is None: 
        decorators = [] 

    #is obj type is part of the default_integration types then skip
    #this prevent infinite recursion during the __init__ call 
    if default_integrations.get(obj.integration_type, {}).get(obj.type, None) is None:
        for integration_type in default_integrations.keys():
            if integration_type not in dependent_integrations.keys(): 
                dependent_integrations[integration_type]={}
            integrations = dependent_integrations.get(integration_type)
  
            for integration in default_integrations.get(integration_type).keys(): 
                #ignore module name
                #skip if we already have the integration
                if integration not in integrations.keys(): 
                    int_obj = register_integration_class(obj, conf, integration, 
                                                        integration_type=integration_type,
                                                        integration_type_conf_name=integration_type, 
                                                        default_class_name=default_integrations.get(integration_type).get(integration), 
                                                        parent_integration_type=obj.integration_type,
                                                        problem_type=obj.problem_type, 
                                                        dependent_integrations=dependent_integrations,
                                                        plugin_mods=plugin_mods,
                                                        decorators=decorators,
                                                        skip_config_validation=skip_config_validation,
                                                        )
                    if int_obj is not None: 
                        dependent_integrations.get(integration_type).update({integration: int_obj})
                        obj.log("Loaded class %s into %s %s" %
                                (int_obj.name, integration_type, integration))
                    else: 
                        raise Exception("Unable to find %s class" %integration)

                else: 
                    setattr(obj, integration, integrations[integration])

    return dependent_integrations

def parse_dict_string(dict_string): 
    return json.loads(dict_string.replace("\'", "\"").replace("False", "false").replace("True", "true"))

def convert_arg_to_string(arg):
    basic_types = (str, bool, int, float, complex) 
    sequence_types = (list, tuple, range, set, listconfig.ListConfig)
    dict_types = (dict, dictconfig.DictConfig)

    if isinstance(arg, basic_types): 
        out = str(arg)
    elif isinstance(arg, sequence_types): 
        out = "" 
        for i in arg:
            out = f"{out}_{convert_arg_to_string(i)}"
        out = out[1:] 
    elif isinstance(arg, dict_types): 
        out = "" 
        for k,v in arg.items(): 
            out = f"{out}_{k}-{convert_arg_to_string(v)}"
    elif arg is None: 
        out = "None"
    else: 
        try: 
            out = arg.__module__
        except: 
            out = "obj"
    
    return out

def compare_objects(objA, objB): 
    basic_types = (str, bool, int, float, complex) 
    sequence_types = (list, tuple, range, set, listconfig.ListConfig)
    dict_types = (dict, dictconfig.DictConfig)

    if objA is not None and objB is not None:
        if callable(objA):
            objA = getsource(objA)
        if callable(objB):
            objB = getsource(objB)

        if isinstance(objA, type(objB)):
            if isinstance(objA, basic_types): 
                return objA == objB 
            elif isinstance(objA, sequence_types): 
                return all([compare_objects(A,B) for A,B in zip(objA, objB)])
            elif isinstance(objA, dict_types): 
                return all([compare_objects(objA.get(i,None),objB.get(i,None)) for i in set(objA.keys()).union(objB.keys())])
            else: 
                if hasattr(objA, "equals"):
                    return objA.equals(objB)
                elif hasattr(objA, "equal"): 
                    return objA.equal(objB)
                else: 
                    try: 
                        return all(list(objA == objB)) 
                    except: 
                        return False
        else:
            return False

    elif objA is None and objB is None: 
        return True 
    else: 
        return False 

def get_docker_string(image_str): 
    return image_str.lower().replace("_","-")


from lolpop.runner.classification_runner.classification_runner import ClassificationRunner
from lolpop.utils import common_utils as utils


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class RegressionRunner(ClassificationRunner):

    def __init__(self, problem_type="regression", *args, **kwargs):
        super().__init__(problem_type=problem_type, *args, **kwargs)


from lolpop.pipeline.train.base_train import BaseTrain
from lolpop.utils import common_utils as utils
from lolpop.utils import metaflow_utils as meta_utils
from metaflow import FlowSpec, step
from pathlib import Path

METAFLOW_CLASS = "MetaflowOfflineTrainSpec"
PLUGIN_PATHS = "plugin_paths.txt"


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class MetaflowOfflineTrain(BaseTrain):
    __REQUIRED_CONF__ = {
        "component": ["data_splitter", "metadata_tracker", "model_checker", "model_explainer", "model_visualizer", "model_bias_checker"],
        "config": []
    }

    def run(self, data, *args, **kwargs):
        """
        This method runs the Metaflow pipeline with the provided data.

        Args:
            data (pandas dataframe): The data to be used in the training process.
            **kwargs (dictionary): Additional keyword arguments that may be needed for the training process.

        Returns:
            None
        """

        #get flow class object from this file
        mod_cl = meta_utils.get_flow_class(__file__, METAFLOW_CLASS)

        flow = meta_utils.load_flow(
            mod_cl, self, PLUGIN_PATHS, data=data)
        self.log("Loaded metaflow flow %s" % METAFLOW_CLASS)

        meta_utils.run_flow(flow, "run", __file__, PLUGIN_PATHS)
        self.log("Metaflow pipeline %s finished." % METAFLOW_CLASS)

    def get_artifacts(self, artifact_keys):
        """
        This method retrieves the artifacts associated with the Metaflow pipeline run.

        Args:
            artifact_keys (list): A list of artifact keys (strings) to retrieve from the run.

        Returns:
            artifacts (dictionary): A dictionary containing the artifacts retrieved from the run.
        """
        #get latest run of this pipeline
        run = meta_utils.get_latest_run(METAFLOW_CLASS)

        #get requested artifacts
        artifacts = meta_utils.get_run_artifacts(
            run, artifact_keys, METAFLOW_CLASS)

        return artifacts


class MetaflowOfflineTrainSpec(FlowSpec):

    def __init__(self, lolpop=None, data=None, use_cli=False, **kwargs):
        #you need to set local attributes before calling super
        #only the first time we create the class will these parameters be provided.
        #The rest of the calls are metaflow internal calls and we do not want to reset them
        if lolpop is not None:
            self.lolpop = lolpop
        if data is not None:
            self.data = data

        #load plugins
        meta_utils.load_plugins(PLUGIN_PATHS)

        #call init -- this kicks off the main workflow
        super().__init__(use_cli=use_cli)

    @step
    def start(self):
        self.next(self.split_data)

    @step
    def split_data(self):
        #split data. returns dictionary of train/vali/test dataframes
        data_dict = self.lolpop.data_splitter.split_data(self.data)
        self.data_dict = data_dict

        self.next(self.train_model)

    @step
    def train_model(self):
        #create model_version
        id = self.lolpop._get_config("model_name")
        model_version = self.lolpop.metadata_tracker.create_resource(id, type="model_version")

        #if we are using hyperparameter tuner, use that, otherwise just use the model trainer provided
        if hasattr(self.lolpop, "hyperparameter_tuner"):
            model = self.lolpop.hyperparameter_tuner.run_experiment(
                self.data_dict, model_version)
        else:
            model = self.lolpop.model_trainer.build_model(self.data_dict, model_version)

        self.model = model 
        self.model_version = model_version 

        self.next(self.analyze_model)

    @step
    def analyze_model(self):

        transformed_data = self.model._transform_dict(self.data_dict)

        #calculate feature importance
        if not (self.lolpop.problem_type == "timeseries"): 
            self.lolpop.model_explainer.get_feature_importance(
                transformed_data, self.model, self.model_version)

        #compare model to baseline
        self.lolpop.model_checker.get_baseline_comparison(
            self.data_dict, self.model, self.model_version)

        #create some eye candy
        self.lolpop.model_visualizer.generate_viz(
            transformed_data, self.model._get_model(), self.model_version)
        
        if (self.lolpop.problem_type == "timeseries"):
            self.lolpop.model_visualizer.cross_validation(
                self.model._get_model(), 
                self.model_version,
                initial=self.model._get_config("cv_initial"),
                period=self.model._get_config("cv_period"),
                horizon=self.model._get_config("cv_horizon"),
            )

        self.next(self.check_model)

    @step
    def check_model(self):
        #don't check model if it's a ts forecasting problem
        if not (self.lolpop.problem_type == "timeseries"):
            #run model checks
            model_report, file_path, checks_status = self.lolpop.model_checker.check_model(
                self.data_dict, self.model)

            #log model report to metadata tracker
            self.lolpop.metadata_tracker.log_checks(
                self.model_version,
                file_path=file_path,
                report=model_report,
                checker_class=self.lolpop.model_checker.name,
                type="model"
            )

            if checks_status == "ERROR" or checks_status == "WARN":
                url = self.lolpop.metadata_tracker.url
                self.lolpop.notify(
                    "Issues found with model checks. Visit %s for more information." % url, checks_status)

        self.next(self.check_model_bias)

    @step
    def check_model_bias(self):
        #don't check bias for ts forecasting
        if not (self.lolpop.problem_type == "timeseries"):
            #check model bias
            bias_metrics = self.lolpop.model_bias_checker.check_model_bias(
                self.data_dict, self.model)
            for k, v in bias_metrics.items():
                self.lolpop.metrics_tracker.log_metric(self.model_version, k, v)

        self.next(self.compare_models)

    @step
    def compare_models(self):
        #don't comapre models for ts forecasting
        if not (self.lolpop.problem_type == "timeseries"):
            #get currently deployed model version/previous model version and winning experiment
            prev_model_version = self.lolpop.metadata_tracker.get_currently_deployed_model_version(
                self.model_version)

            if prev_model_version is not None:
                prev_experiment = self.lolpop.metadata_tracker.get_winning_experiment(
                    prev_model_version)

                #get the prev model version object from resource version control and load it into a model trainer
                prev_model_obj = self.lolpop.resource_version_control.get_model(
                    prev_experiment)
                prev_model = self.lolpop.metadata_tracker.load_model(
                    prev_model_obj, self.model_version, self.model)

                #compare model to the last version
                report, file_path = self.lolpop.model_checker.calculate_model_drift(
                    self.data_dict, self.model, prev_model)
                self.lolpop.metadata_tracker.log_artifact(
                    self.model_version, id="model_drift_report", path=file_path, external=False)
                is_new_model_better = self.lolpop.model_checker.compare_models(
                    self.data_dict, self.model, prev_model, self.model_version, prev_model_version)

            else:
                self.lolpop.log("No previous model version found for model: %s" %
                        self.lolpop.metadata_tracker.get_resource_id(self.model_version))
                is_new_model_better = True 
        else: 
            is_new_model_better = True     

        self.is_new_model_better = is_new_model_better
          
        self.next(self.retrain_model_on_all_data)


    @step
    def retrain_model_on_all_data(self):

        #only retrain if the new model is better than last model and 
        #retrain_all config is set
        if self.is_new_model_better: 
            if self.lolpop._get_config("retrain_all"): 
                if hasattr(self.lolpop, "model_trainer"):
                    model, experiment = self.lolpop.model_trainer.rebuild_model(
                        self.data_dict, self.model_version)
                else:
                    #if no model_trainer is specified in config, we can rebuild using the winning_exp_model_trainer
                    winning_exp_model = self.lolpop.metadata_tracker.load_model(
                        None, self.model_version, ref_model=self.model)
                    model, experiment = winning_exp_model.rebuild_model(
                        self.data_dict, self.model_version)
                self.model = model 

        self.next(self.end)

    @step
    def end(self):
        pass


if __name__ == '__main__':
    #need to use CLI here because metaflow calls back into this function to
    #execute each individual step
    MetaflowOfflineTrainSpec(use_cli=True)



from lolpop.pipeline.base_pipeline import BasePipeline
from lolpop.utils import common_utils as utils
from typing import Any 

class BaseTrain(BasePipeline): 
    def __init__(self, pipeline_type = "train", *args, **kwargs):
        super().__init__(pipeline_type = pipeline_type, *args, **kwargs)

    def split_data(self, data, *args, **kwargs) -> Any: 
        pass 

    def train_model(self, data, *args, **kwargs) -> tuple[Any, Any]: 
        pass 

    def check_model(self, data, model, model_version, *args, **kwargs):
        pass

    def analyze_model(self, data, model, model_verison, *args, **kwargs):
        pass

    def compare_models(self, data, model, model_version, *args, **kwargs) -> bool:
        pass

    def check_model_bias(self, data, model, model_version, *args, **kwargs):
        pass

    def retrain_model_on_all_data(self, data, model_version, *args, **kwargs) -> tuple[Any, Any]: 
        pass 



from lolpop.component.genai_chatbot.base_genai_chatbot import BaseGenAIChatbot
from lolpop.utils import common_utils as utils

import openai

class OpenAIChatbot(BaseGenAIChatbot):

    __REQUIRED_CONF__= {"config": ["OPENAI_API_KEY"]}

    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs)
        self.api_key = utils.load_config(["OPENAI_API_KEY"], self.config).get("OPENAI_API_KEY")

    def ask(self, messages=None, model="gpt-3.5-turbo", *args, **kwargs):
        """
        Sends a chat message(s) to OpenAI's Chat API and returns the response.

        Args:
            messages (list): A list of strings representing the messages to send to the chatbot. Default is an empty list.
            model (str): A string representing the name of the model to use for generating chat responses. Default is "gpt-3.5-turbo".
            **kwargs: Arbitrary keyword arguments representing the additional options to pass to the Chat API.

        Returns:
            A string representing the generated chatbot response.

        Raises:
            Exception: If no messages are provided to the chatbot.
        """

        if messages is None or len(messages) == 0: 
            raise Exception("Error: No messages provided to chatbot. Please try again with non-null content.")
        else: 
            completion = openai.ChatCompletion.create(
                model=model,
                messages=messages,
                **kwargs
            )

            return completion.choices[0].message.content
        
    def prepare_message(self, role="user", content=None, *args, **kwargs): 
        """
        Prepares a chat message with the specified role and content.

        Args:
            role (str): A string representing the role of the person sending the message. Default is "user".
            content (str): A string representing the content of the message to be sent. Default is None.

        Returns:
            A dictionary representing the prepared message with the specified role and content.
        """
        return {"role": role, "content": content}



from lolpop.component.base_component import BaseComponent
from omegaconf.listconfig import ListConfig
from lolpop.utils import common_utils as utils
import itertools 

from typing import Any 

class BaseHyperparameterTuner(BaseComponent): 
    __REQUIRED_CONF__ = {
        "component": ["metadata_tracker", "resource_version_control", "metrics_tracker"],
        "config" : ["training_params", "metrics", "perf_metric"],
    }
    
    def run_experiment(self, data, *args, **kwargs) -> Any: 
        pass 

    def build_model(self, data, model_version, algo, params, trainer_config = None, *args, **kwargs) -> tuple[Any, Any]: 
        if trainer_config is None: 
            trainer_config = {}
        #create experiment and log params
        experiment = self.metadata_tracker.create_resource(id=None, type="experiment", parent=model_version)

        #load model trainer and build model
        model_cl = utils.load_class(algo)
        dependent_integrations = {"component": {"logger": self.logger, "notifier": self.notifier,  "metadata_tracker": self.metadata_tracker,
                                  "metrics_tracker": self.metrics_tracker, "resource_version_control": self.resource_version_control}}
        if hasattr(self, "feature_transformer"): #pass the feature_transformer if it's defined at the pipeline level 
            dependent_integrations.get("component").update({"feature_transformer": self.feature_transformer})
        model = model_cl(conf=trainer_config, 
                         parent=self,
                         is_standalone=True,  
                         problem_type=self.problem_type, 
                         params=params, 
                         dependent_integrations=dependent_integrations)

        #now fit model 
        model_obj = model.transform_and_fit(data)
        
        #save model 
        model.save(experiment)

        return model, experiment 

    def save_model(self, model, experiment, *args, **kwargs):
        #save model
        model.save(experiment)

    #builds a grid of all possible parameter combinations gives a params item with a list of values for each param
    def _build_training_grid(self, params) -> dict[str, Any]: 
        keys, values = zip(*params.items())
        values = [v if isinstance(v, ListConfig) else str(v).split(",") for v in values]
        grid = [dict(zip(keys,v)) for v in itertools.product(*values)]

        return grid

    #determines the winning experiment from the experiment list 
    def _get_winning_experiment(self, exp_list, reverse=False) -> Any: 
        # sort list by values. reverse=True is descending, False is ascending (Default)
        new_list = sorted(exp_list.items(), key=lambda x: x[1], reverse=reverse) 
        return new_list[0][0]


from lolpop.component.model_deployer.base_model_deployer import BaseModelDeployer
from lolpop.utils import common_utils as utils

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class SeldonModelDeployer(BaseModelDeployer): 
    __REQUIRED_CONF__ = {
        "config" : []
    }
    def deploy_model(self, model, *args, **kwargs): 
        pass 

    def build_api(self, model, *args, **kwargs): 
        pass 

    def build_container(self, model, *args, **kwargs): 
        pass 

    def package_model(self, model, *args, **kwargs): 
        pass


from lolpop.component.base_component import BaseComponent
from typing import Any

class BaseDataChecker(BaseComponent): 

    def check_data(self, data, *args, **kwargs) -> tuple[Any, str, str]: 
        pass 




from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

import pandas as pd
import duckdb

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class DuckDBDataConnector(BaseDataConnector):
    __REQUIRED_CONF__ = {"config": ["duckdb_path"]}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = self._get_config("duckdb_path")

    def get_data(self, table, sql=None, *args, **kwargs):
        """
        Retrieves data from the DuckDBDataConnector table or custom SQL provided and returns in a pandas dataframe.

        Args:
            table (str): Name of the table to retrieve data from.
            sql (str): The optional SQL query to execute. Default value is None.

        Returns:
            data: Pandas dataframe object containing the data.

        Raises:
            Exception: If both table and sql statement are not provided.
        """
        if sql is None:
            if table is not None:
                sql = "SELECT * FROM %s" % table
            else:
                raise Exception(
                    "Table or SQL command must be provided to DuckDBDataConnector in order to retrieve data.")

        data = self._load_data(sql, self.path)

        return data

    def save_data(self, data, table, *args, **kwargs):
        """
        Saves data to the specified table in the specified DuckDBDataConnector instance.
        If the table does not exist, it gets created with the data structure from the dataframe provided.
        If a column is missing, it adds the column as nulls. This preserves the structure of the destination table.

        Args:
            data (pandas.DataFrame): Pandas dataframe containing the data to be saved.
            table (str): Name of the table to save the data to.

        """
        #check if table already exists
        tables = self.get_data(None, sql="SHOW TABLES")

        #if table exists, see if we've added new columns or deleted columns
        table_exists = tables[tables["name"]==table].shape[0] > 0
        if table_exists:
           table_schema = self.get_data(table)
           new_features = [x for x in data.columns if x.upper(
           ) not in table_schema.columns.str.upper()]
           deleted_features = [
               x for x in table_schema.columns if x.upper() not in data.columns.str.upper()]

           #if we have new columns, we need to alter the table definiton
           if len(new_features) > 0:
               sql = "ALTER TABLE %s " % (table)
               for col_name in new_features:
                    col_type = data.dtypes[col_name]
                    duckdb_type = self.__map_pandas_col_type_to_duckdb_type(col_type)
                    sql = sql + ("ADD COLUMN %s %s" %
                                 (col_name.upper(), duckdb_type))
               self.get_data(None, sql=sql)

           #if we've deleted any features, then just add blank rows to the dataframe
           if len(deleted_features) > 0:
               for feature in deleted_features:
                   data[feature] = None

        #now we can save our data
        self._save_data(data, table, self.path, table_exists)

     #load data into df
    @classmethod
    def _load_data(self, sql, path, *args, **kwargs):
        """
        Loads data into a pandas dataframe.
        
        Args:
            sql (str): SQL query to execute to fetch data.
            path (str): Path to database.
        
        Returns:
            result: Pandas dataframe object containing the data from the query.
        """
        result = pd.DataFrame()
        with duckdb.connect(database = path) as con: 
            res = con.sql(sql)
            result = res.df() 
        return result

    def _save_data(self, data, table_name, path, table_exists, *args, **kwargs):
        """
        Saves data into the DuckDBDataConnector table.
        
        Args:
            data (pd.DataFrame): Pandas dataframe object containing data.
            table_name (str): Name of the table.
            path (str): Path to database.
            table_exists (bool): Check if the table exists or not.
        
        Returns: 
            None

        """
        if data is not None and len(data)>0: 
            with duckdb.connect(database = path) as con: 
                if not table_exists: 
                    con.sql("CREATE TABLE %s as SELECT * from data" %table_name)
                else: 
                    con.sql("INSERT INTO %s SELECT * from data" % table_name)


    def __map_pandas_col_type_to_duckdb_type(self, col_type):
        """
        This is a private method. It maps pandas data types to duckdb data types.
        
        Args:
            col_type (pd.dtype): Pandas data type.

        Returns:
            column_type: DuckDB data type corresponding to pandas data type.

        """
        if col_type.kind == 'M':
            # datetime and timestamp columns in pandas are converted to datetime64[ns] dtype, which corresponds to 'TIMESTAMP'
            column_type = 'TIMESTAMP'
        elif col_type.kind == 'm':
            # timedelta columns in pandas are converted to timedelta64[ns] dtype, which corresponds to 'INTERVAL'
            column_type = 'INTERVAL'
        elif col_type.kind == 'b':
            # boolean columns in pandas are converted to 'bool' dtype, which corresponds to 'boolean'
            column_type = 'BOOLEAN'
        elif col_type.kind == 'i': 
            column_type = 'BIGINT'
        elif col_type.kind == 'u':
            column_type = 'UBIGINT'
        elif col_type.kind == 'f' or col_type.kind == 'c':
            # float columns in pandas are converted to 'float64' dtype, which corresponds to 'float'
            column_type = 'FLOAT'
        else:
            # All other column types in pandas are assumed to be string columns, which correspond to 'varchar' in Snowflake
            column_type = 'VARCHAR'
        return column_type



from lolpop.component.base_component import BaseComponent
from typing import Any

class BaseDataSplitter(BaseComponent): 

    def split_data(self, data, *args, **kwargs) -> dict[str, Any]: 
        pass 

    def get_train_test_dfs(self, data, *args, **kwargs) -> tuple[Any, Any]: 
        pass 
